#!/usr/bin/env python3
#
#   test_01_log.py (C) 2021-2025, Peter Sulyok
#   Unit test for smfc.Log() class.
#
import syslog
import pytest
from mock import MagicMock
from pytest_mock import MockerFixture
from smfc.log import Log


class TestLog:
    """Unit test class for smfc.Log() class"""

    @pytest.mark.parametrize("level, output, error", [
        (Log.LOG_NONE, Log.LOG_STDOUT, "Log.__init__() 1"),
        (Log.LOG_ERROR, Log.LOG_STDOUT, "Log.__init__() 2"),
        (Log.LOG_CONFIG, Log.LOG_STDOUT, "Log.__init__() 3"),
        (Log.LOG_INFO, Log.LOG_STDOUT, "Log.__init__() 4"),
        (Log.LOG_DEBUG, Log.LOG_STDOUT, "Log.__init__() 5"),
        (Log.LOG_NONE, Log.LOG_STDERR, "Log.__init__() 6"),
        (Log.LOG_ERROR, Log.LOG_STDERR, "Log.__init__() 7"),
        (Log.LOG_CONFIG, Log.LOG_STDERR, "Log.__init__() 8"),
        (Log.LOG_INFO, Log.LOG_STDERR, "Log.__init__() 9"),
        (Log.LOG_DEBUG, Log.LOG_STDERR, "Log.__init__() 10"),
        (Log.LOG_NONE, Log.LOG_SYSLOG, "Log.__init__() 11"),
        (Log.LOG_ERROR, Log.LOG_SYSLOG, "Log.__init__() 12"),
        (Log.LOG_CONFIG, Log.LOG_SYSLOG, "Log.__init__() 13"),
        (Log.LOG_INFO, Log.LOG_SYSLOG, "Log.__init__() 14"),
        (Log.LOG_DEBUG, Log.LOG_SYSLOG, "Log.__init__() 15")
    ])
    def test_init_p1(self, mocker:MockerFixture, level: int, output: int, error: str) -> None:
        """Positive unit test for Log.__init__() method. It contains the following steps:
            - mock print(), syslog.openlog(), and syslog.syslog() functions
            - initialize a Log class instance with a specified level and output values
            - ASSERT: if the class attributes contain different values that were passed to __init__
            - ASSERT: if the mocked system functions were called wrong number of times
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_syslog_openlog = MagicMock()
        mocker.patch('syslog.openlog', mock_syslog_openlog)
        mock_syslog_syslog = MagicMock()
        mocker.patch('syslog.syslog', mock_syslog_syslog)
        my_log = Log(level, output)
        assert my_log.log_level == level, error
        assert my_log.log_output == output, error
        if my_log.log_output is Log.LOG_STDOUT:
            assert my_log.msg == my_log.msg_to_stdout, error
        elif my_log.log_output is Log.LOG_STDERR:
            assert my_log.msg == my_log.msg_to_stderr, error
        elif my_log.log_output == Log.LOG_SYSLOG:
            assert my_log.msg == my_log.msg_to_syslog, error

    @pytest.mark.parametrize("level, output, error", [
        (100, Log.LOG_STDOUT, "Log.__init__() 16"),
        (Log.LOG_ERROR, 100, "Log.__init__() 17")
    ])
    def test_init_n1(self, mocker: MockerFixture, level: int, output: int, error: str) -> None:
        """Negative unit test for Log.__init__() method. It contains the following steps:
            - initialize a Log class instance with specified level and output values
            - mock print() function
            - ASSERT: if __init__ does not raise an exception in case of invalid value
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        with pytest.raises(ValueError) as cm:
            Log(level, output)
        assert cm.type is ValueError, error

    @pytest.mark.parametrize("level, syslog_level, error", [
        (Log.LOG_ERROR, syslog.LOG_ERR, "Log.map_to_syslog() 1"),
        (Log.LOG_CONFIG, syslog.LOG_INFO, "Log.map_to_syslog() 2"),
        (Log.LOG_INFO, syslog.LOG_INFO, "Log.map_to_syslog() 3"),
        (Log.LOG_DEBUG, syslog.LOG_DEBUG, "Log.map_to_syslog() 4"),
        (1000, syslog.LOG_ERR, "Log.map_to_syslog() 5"),
        (-1, syslog.LOG_ERR, "Log.map_to_syslog() 5")
    ])
    def test_mts(self, level: int, syslog_level: int, error: str) -> None:
        """Positive unit test for Log.map_to_syslog() method. It contains the following steps:
            - initialize a Log class instance
            - call map_to_syslog with a specified level
            - ASSERT: if map_to_syslog function maps a specified log levels to a wrong syslog level value
        """
        assert Log.map_to_syslog(level) == syslog_level, error

    @pytest.mark.parametrize("level, level_str, error", [
        (Log.LOG_NONE, 'NONE', "Log.level_to_str() 1"),
        (Log.LOG_ERROR, 'ERROR', "Log.level_to_str() 2"),
        (Log.LOG_CONFIG, 'CONFIG', "Log.level_to_str() 3"),
        (Log.LOG_INFO, 'INFO', "Log.level_to_str() 4"),
        (Log.LOG_DEBUG, 'DEBUG', "Log.level_to_str() 5"),
        (-1, 'NONE', "Log.level_to_str() 6"),
        (1000, 'NONE', "Log.level_to_str() 7")
    ])
    def test_lts(self, level: int, level_str: str, error: str) -> None:
        """Positive unit test for Log.level_to_str() method. It contains the following steps:
            - call level_to_str() with a specified level
            - ASSERT: if level_to_str function maps a log level to an invalid string value
        """
        assert Log.level_to_str(level) == level_str, error

    @pytest.mark.parametrize("level, output, msg_level, count, error", [
        (Log.LOG_NONE, Log.LOG_STDOUT, Log.LOG_NONE, 0, "Log.msg_to_xxx() 1"),
        (Log.LOG_NONE, Log.LOG_STDOUT, Log.LOG_ERROR, 0, "Log.msg_to_xxx() 2"),
        (Log.LOG_NONE, Log.LOG_STDOUT, Log.LOG_CONFIG, 0, "Log.msg_to_xxx() 3"),
        (Log.LOG_NONE, Log.LOG_STDOUT, Log.LOG_INFO, 0, "Log.msg_to_xxx() 3"),
        (Log.LOG_NONE, Log.LOG_STDOUT, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 4"),

        (Log.LOG_NONE, Log.LOG_STDERR, Log.LOG_NONE, 0, "Log.msg_to_xxx() 5"),
        (Log.LOG_NONE, Log.LOG_STDERR, Log.LOG_ERROR, 0, "Log.msg_to_xxx() 6"),
        (Log.LOG_NONE, Log.LOG_STDERR, Log.LOG_CONFIG, 0, "Log.msg_to_xxx() 7"),
        (Log.LOG_NONE, Log.LOG_STDERR, Log.LOG_INFO, 0, "Log.msg_to_xxx() 8"),
        (Log.LOG_NONE, Log.LOG_STDERR, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 9"),

        (Log.LOG_NONE, Log.LOG_SYSLOG, Log.LOG_NONE, 0, "Log.msg_to_xxx() 10"),
        (Log.LOG_NONE, Log.LOG_SYSLOG, Log.LOG_ERROR, 0, "Log.msg_to_xxx() 11"),
        (Log.LOG_NONE, Log.LOG_SYSLOG, Log.LOG_CONFIG, 0, "Log.msg_to_xxx() 12"),
        (Log.LOG_NONE, Log.LOG_SYSLOG, Log.LOG_INFO, 0, "Log.msg_to_xxx() 13"),
        (Log.LOG_NONE, Log.LOG_SYSLOG, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 14"),

        (Log.LOG_ERROR, Log.LOG_STDOUT, Log.LOG_NONE, 0, "Log.msg_to_xxx() 15"),
        (Log.LOG_ERROR, Log.LOG_STDOUT, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 16"),
        (Log.LOG_ERROR, Log.LOG_STDOUT, Log.LOG_CONFIG, 0, "Log.msg_to_xxx() 17"),
        (Log.LOG_ERROR, Log.LOG_STDOUT, Log.LOG_INFO, 0, "Log.msg_to_xxx() 18"),
        (Log.LOG_ERROR, Log.LOG_STDOUT, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 19"),

        (Log.LOG_ERROR, Log.LOG_STDERR, Log.LOG_NONE, 0, "Log.msg_to_xxx() 20"),
        (Log.LOG_ERROR, Log.LOG_STDERR, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 21"),
        (Log.LOG_ERROR, Log.LOG_STDERR, Log.LOG_CONFIG, 0, "Log.msg_to_xxx() 22"),
        (Log.LOG_ERROR, Log.LOG_STDERR, Log.LOG_INFO, 0, "Log.msg_to_xxx() 23"),
        (Log.LOG_ERROR, Log.LOG_STDERR, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 24"),

        (Log.LOG_ERROR, Log.LOG_SYSLOG, Log.LOG_NONE, 0, "Log.msg_to_xxx() 25"),
        (Log.LOG_ERROR, Log.LOG_SYSLOG, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 26"),
        (Log.LOG_ERROR, Log.LOG_SYSLOG, Log.LOG_CONFIG, 0, "Log.msg_to_xxx() 27"),
        (Log.LOG_ERROR, Log.LOG_SYSLOG, Log.LOG_INFO, 0, "Log.msg_to_xxx() 28"),
        (Log.LOG_ERROR, Log.LOG_SYSLOG, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 29"),

        (Log.LOG_CONFIG, Log.LOG_STDOUT, Log.LOG_NONE, 0, "Log.msg_to_xxx() 30"),
        (Log.LOG_CONFIG, Log.LOG_STDOUT, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 31"),
        (Log.LOG_CONFIG, Log.LOG_STDOUT, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 32"),
        (Log.LOG_CONFIG, Log.LOG_STDOUT, Log.LOG_INFO, 0, "Log.msg_to_xxx() 33"),
        (Log.LOG_CONFIG, Log.LOG_STDOUT, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 34"),

        (Log.LOG_CONFIG, Log.LOG_STDERR, Log.LOG_NONE, 0, "Log.msg_to_xxx() 35"),
        (Log.LOG_CONFIG, Log.LOG_STDERR, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 36"),
        (Log.LOG_CONFIG, Log.LOG_STDERR, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 37"),
        (Log.LOG_CONFIG, Log.LOG_STDERR, Log.LOG_INFO, 0, "Log.msg_to_xxx() 38"),
        (Log.LOG_CONFIG, Log.LOG_STDERR, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 39"),

        (Log.LOG_CONFIG, Log.LOG_SYSLOG, Log.LOG_NONE, 0, "Log.msg_to_xxx() 40"),
        (Log.LOG_CONFIG, Log.LOG_SYSLOG, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 41"),
        (Log.LOG_CONFIG, Log.LOG_SYSLOG, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 42"),
        (Log.LOG_CONFIG, Log.LOG_SYSLOG, Log.LOG_INFO, 0, "Log.msg_to_xxx() 43"),
        (Log.LOG_CONFIG, Log.LOG_SYSLOG, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 44"),

        (Log.LOG_INFO, Log.LOG_STDOUT, Log.LOG_NONE, 0, "Log.msg_to_xxx() 45"),
        (Log.LOG_INFO, Log.LOG_STDOUT, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 46"),
        (Log.LOG_INFO, Log.LOG_STDOUT, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 47"),
        (Log.LOG_INFO, Log.LOG_STDOUT, Log.LOG_INFO, 1, "Log.msg_to_xxx() 48"),
        (Log.LOG_INFO, Log.LOG_STDOUT, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 49"),

        (Log.LOG_INFO, Log.LOG_STDERR, Log.LOG_NONE, 0, "Log.msg_to_xxx() 50"),
        (Log.LOG_INFO, Log.LOG_STDERR, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 51"),
        (Log.LOG_INFO, Log.LOG_STDERR, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 52"),
        (Log.LOG_INFO, Log.LOG_STDERR, Log.LOG_INFO, 1, "Log.msg_to_xxx() 53"),
        (Log.LOG_INFO, Log.LOG_STDERR, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 54"),

        (Log.LOG_INFO, Log.LOG_SYSLOG, Log.LOG_NONE, 0, "Log.msg_to_xxx() 55"),
        (Log.LOG_INFO, Log.LOG_SYSLOG, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 56"),
        (Log.LOG_INFO, Log.LOG_SYSLOG, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 57"),
        (Log.LOG_INFO, Log.LOG_SYSLOG, Log.LOG_INFO, 1, "Log.msg_to_xxx() 58"),
        (Log.LOG_INFO, Log.LOG_SYSLOG, Log.LOG_DEBUG, 0, "Log.msg_to_xxx() 59"),

        (Log.LOG_DEBUG, Log.LOG_STDOUT, Log.LOG_NONE, 0, "Log.msg_to_xxx() 60"),
        (Log.LOG_DEBUG, Log.LOG_STDOUT, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 61"),
        (Log.LOG_DEBUG, Log.LOG_STDOUT, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 62"),
        (Log.LOG_DEBUG, Log.LOG_STDOUT, Log.LOG_INFO, 1, "Log.msg_to_xxx() 63"),
        (Log.LOG_DEBUG, Log.LOG_STDOUT, Log.LOG_DEBUG, 1, "Log.msg_to_xxx() 64"),

        (Log.LOG_DEBUG, Log.LOG_STDERR, Log.LOG_NONE, 0, "Log.msg_to_xxx() 65"),
        (Log.LOG_DEBUG, Log.LOG_STDERR, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 66"),
        (Log.LOG_DEBUG, Log.LOG_STDERR, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 67"),
        (Log.LOG_DEBUG, Log.LOG_STDERR, Log.LOG_INFO, 1, "Log.msg_to_xxx() 68"),
        (Log.LOG_DEBUG, Log.LOG_STDERR, Log.LOG_DEBUG, 1, "Log.msg_to_xxx() 69"),

        (Log.LOG_DEBUG, Log.LOG_SYSLOG, Log.LOG_NONE, 0, "Log.msg_to_xxx() 70"),
        (Log.LOG_DEBUG, Log.LOG_SYSLOG, Log.LOG_ERROR, 1, "Log.msg_to_xxx() 71"),
        (Log.LOG_DEBUG, Log.LOG_SYSLOG, Log.LOG_CONFIG, 1, "Log.msg_to_xxx() 72"),
        (Log.LOG_DEBUG, Log.LOG_SYSLOG, Log.LOG_INFO, 1, "Log.msg_to_xxx() 73"),
        (Log.LOG_DEBUG, Log.LOG_SYSLOG, Log.LOG_DEBUG, 1, "Log.msg_to_xxx() 74")
    ])
    def test_msg_to_xxx(self, mocker:MockerFixture, level: int, output: int, msg_level: int, count: int,
                        error: str) -> None:
        """Positive unit test for Log.msg_to_xxx() method. It contains the following steps:
            - mock print() and syslog.syslog() functions
            - initialize a Log class instance with a specified level and output
            - calls Log.msg() with a specified level and log message
            - ASSERT: if msg() function calls print and/or syslog functions other times than expected
              (in case of DEBUG level there are additional 3 calls)
            - delete the Log instance
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_syslog_openlog = MagicMock()
        mocker.patch('syslog.openlog', mock_syslog_openlog)
        mock_syslog_syslog = MagicMock()
        mocker.patch('syslog.syslog', mock_syslog_syslog)
        my_log = Log(level, output)
        my_log.msg(msg_level, "This is a test log message.")
        if output == Log.LOG_STDOUT:
            assert mock_print.call_count == count, error
        elif output == Log.LOG_STDERR:
            assert mock_print.call_count == count, error
        elif output == Log.LOG_SYSLOG:
            assert mock_syslog_syslog.call_count == count, error

# End.



#
#   service.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.Service() class implementation.
#
import atexit
import os
import sys
import time
from importlib.metadata import version
from configparser import ConfigParser
from argparse import ArgumentParser, Namespace
from pyudev import Context
from smfc.gpuzone import GpuZone
from smfc.cpuzone import CpuZone
from smfc.hdzone import HdZone
from smfc.ipmi import Ipmi
from smfc.log import Log


class Service:
    """Service class contains all resources/functions for the execution."""

    # Service data.
    config: ConfigParser    # Instance for a parsed configuration
    sudo: bool              # Use sudo command
    log: Log                # Instance for a Log class
    udevc: Context          # Reference to a pyudev Context instance (i.e. udev database connection)
    ipmi: Ipmi              # Instance for an Ipmi class
    cpu_zone: CpuZone       # Instance for a CPU Zone fan controller class
    hd_zone: HdZone         # Instance for an HD Zone fan controller class
    gpu_zone: GpuZone       # Instance for an GPU Zone fan controller class
    cpu_zone_enabled: bool  # CPU zone fan controller enabled
    hd_zone_enabled: bool   # HD zone fan controller enabled
    gpu_zone_enabled: bool  # GPU zone fan controller enabled

    def exit_func(self) -> None:
        """This function is called at exit (in case of exceptions or runtime errors cannot be handled), and it switches
           all fans back to rhw default speed 100%, in order to avoid system overheating while `smfc` is not running."""
        # Configure fans.
        if hasattr(self, 'ipmi'):
            self.ipmi.set_fan_level(Ipmi.CPU_ZONE, 100)
            self.ipmi.set_fan_level(Ipmi.HD_ZONE, 100)
            if hasattr(self, 'log'):
                self.log.msg(Log.LOG_INFO, 'smfc terminated: all fans are switched back to the 100% speed.')

        # Unregister this function.
        atexit.unregister(self.exit_func)

    def check_dependencies(self) -> str:
        """Check run-time dependencies of smfc:
              - ipmitool command
              - if CPU zone enabled: `coretemp` or `k10temp` kernel module
              - if HD zone enabled: `drivetemp` kernel module or `smartctl` command
        Returns:
             (str): error string (empty = no errors)

        """
        path: str
        no_smartctl: bool = False
        no_drivetemp: bool = False

        # Check if `ipmitool` command is available.
        path = self.config[Ipmi.CS_IPMI].get(Ipmi.CV_IPMI_COMMAND, '/usr/bin/ipmitool')
        if not os.path.exists(path):
            return f'ERROR: ipmitool command cannot be found {path}!'

        # Load list of kernel modules.
        with open('/proc/modules', 'rt', encoding='utf-8') as file:
            modules = file.read()

        # Check kernel modules for CPUs
        if self.cpu_zone_enabled:
            if 'coretemp' not in modules and 'k10temp' not in modules:
                return 'ERROR: coretemp or k10temp kernel module must be loaded!'

        # Check dependencies for disks
        if self.hd_zone_enabled:

            # Check if `smartctl` command is available.
            path = self.config[HdZone.CS_HD_ZONE].get(HdZone.CV_HD_ZONE_SMARTCTL_PATH, '/usr/sbin/smartctl')
            if not os.path.exists(path):
                no_smartctl = True

            # Check if `drivetemp` modules is loaded.
            if 'drivetemp' not in modules:
                no_drivetemp = True

            # If neither `drivetemp` nor `smartctl` is available.
            if no_smartctl and no_drivetemp:
                return f'ERROR: drivetemp kernel module must be loaded or smartctl command ({path}) must be installed!'

        # If Standby Guard feature enabled, `smartctl` command should be available
        sge = self.config[HdZone.CS_HD_ZONE].getboolean(HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED, fallback=False)
        if sge and no_smartctl:
            return f'ERROR: smartctl command ({path}) must be installed for Standby Guard feature!'

        # All required run-time dependencies are available.
        return ''

    def run(self) -> None:
        """Run function: main execution function of the systemd service.

        Program exit codes:
        0 - printing help or version text (argument parser)
        2 - invalid parameter (argument parser)
        5 - log system initialization error
        6 - config file error
        7 - runtime dependency error
        8 - IPMI initialization error
        9 - udev initialization error
        10 - none of the zones is enabled
        """
        app_parser: ArgumentParser     # Instance for an ArgumentParser class
        parsed_results: Namespace      # Results of parsed command line arguments
        old_mode: int                  # Old IPMI fan mode

        # Handling of the command line arguments.
        app_parser = ArgumentParser()
        # Syntax definition of the command-line parameters.
        app_parser.add_argument('-c', action='store', dest='config_file', default='smfc.conf',
                                help='configuration file')
        app_parser.add_argument('-v', action='version', version='%(prog)s ' + version("smfc"))
        app_parser.add_argument('-l', type=int, choices=[0, 1, 2, 3, 4], default=1,
                                help='log level: 0-NONE, 1-ERROR(default), 2-CONFIG, 3-INFO, 4-DEBUG')
        app_parser.add_argument('-o', type=int, choices=[0, 1, 2], default=2,
                                help='log output: 0-stdout, 1-stderr, 2-syslog(default)')
        app_parser.add_argument('-nd', action='store_true', default=False,
                                help='no dependency checking')
        app_parser.add_argument('-s', action='store_true', default=False,
                                help='use sudo command')
        app_parser.add_argument('-ne', action='store_true', default=False,
                                help='no fan speed recovery at exit')
        # Parsing of the current arguments.
        parsed_results = app_parser.parse_args()

        # Register the emergency exit function for service termination.
        if not parsed_results.ne:
            atexit.register(self.exit_func)

        # Store `sudo` option.
        self.sudo = parsed_results.s

        # Create a Log class instance (in theory, this cannot fail).
        try:
            self.log = Log(parsed_results.l, parsed_results.o)
        except ValueError as e:
            print(f'ERROR: {e}.', flush=True, file=sys.stdout)
            sys.exit(5)

        # Log command line parameters.
        if self.log.log_level >= Log.LOG_CONFIG:
            self.log.msg(Log.LOG_CONFIG, f'Smfc version {version("smfc")} started')
            self.log.msg(Log.LOG_CONFIG, 'Command line arguments:')
            self.log.msg(Log.LOG_CONFIG, f'   original arguments: {" ".join(sys.argv[:])}')
            self.log.msg(Log.LOG_CONFIG, f'   parsed config file = {parsed_results.config_file}')
            self.log.msg(Log.LOG_CONFIG, 'Logging was initialized with:')
            self.log.msg(Log.LOG_CONFIG, f'   log_level = {self.log.log_level}')
            self.log.msg(Log.LOG_CONFIG, f'   log_output = {self.log.log_output}')

        # Parse and load configuration file.
        self.config = ConfigParser()
        if not self.config or not self.config.read(parsed_results.config_file):
            self.log.msg(Log.LOG_ERROR, f'Cannot load configuration file ({parsed_results.config_file})')
            sys.exit(6)
        self.cpu_zone_enabled = self.config[CpuZone.CS_CPU_ZONE].getboolean(CpuZone.CV_CPU_ZONE_ENABLED, fallback=False)
        self.hd_zone_enabled = self.config[HdZone.CS_HD_ZONE].getboolean(HdZone.CV_HD_ZONE_ENABLED, fallback=False)
        self.gpu_zone_enabled = self.config[GpuZone.CS_GPU_ZONE].getboolean(GpuZone.CV_GPU_ZONE_ENABLED, fallback=False)
        self.log.msg(Log.LOG_DEBUG, f'Configuration file ({parsed_results.config_file}) loaded')

        # Check run-time dependencies (commands, kernel modules) if `-nd` command line option is not specified.
        if not parsed_results.nd:
            error_msg = self.check_dependencies()
            if error_msg:
                self.log.msg(Log.LOG_ERROR, error_msg)
                sys.exit(7)

        # Create an Ipmi class instances.
        try:
            self.ipmi = Ipmi(self.log, self.config, self.sudo)
            old_mode = self.ipmi.get_fan_mode()
        except (ValueError, FileNotFoundError) as e:
            self.log.msg(Log.LOG_ERROR, f'{e}.')
            sys.exit(8)
        # Log the old fan mode and zone levels in DEBUG log mode.
        if self.log.log_level >= Log.LOG_DEBUG:
            self.log.msg(Log.LOG_DEBUG, f'Old IPMI fan mode = {self.ipmi.get_fan_mode_name(old_mode)} ({old_mode})')
            self.log.msg(Log.LOG_DEBUG, f'Old CPU zone (0) level = {self.ipmi.get_fan_level(Ipmi.CPU_ZONE)}%')
            self.log.msg(Log.LOG_DEBUG, f'Old HD zone (1) level = {self.ipmi.get_fan_level(Ipmi.HD_ZONE)}%')
        #  Set the FULL IPMI fan mode if it is not the current fan mode.
        if old_mode != Ipmi.FULL_MODE:
            self.ipmi.set_fan_mode(Ipmi.FULL_MODE)
            self.log.msg(Log.LOG_DEBUG,
                         f'New IPMI fan mode = {self.ipmi.get_fan_mode_name(Ipmi.FULL_MODE)}')

        # Initialize connection to udev database
        try:
            self.udevc = Context()
        except ImportError as e:
            self.log.msg(Log.LOG_ERROR, f'pyudev error: Could not interface with libudev: {e}.')
            sys.exit(9)

        # Create an instance for CPU zone fan controller if enabled.
        if self.cpu_zone_enabled:
            self.log.msg(Log.LOG_DEBUG, 'CPU zone fan controller enabled')
            self.cpu_zone = CpuZone(self.log, self.udevc, self.ipmi, self.config)

        # Create an instance for HD zone fan controller if enabled.
        if self.hd_zone_enabled:
            self.log.msg(Log.LOG_DEBUG, 'HD zone fan controller enabled')
            self.hd_zone = HdZone(self.log, self.udevc, self.ipmi, self.config, self.sudo)

        # Create an instance for GPU zone fan controller if enabled.
        if self.gpu_zone_enabled:
            self.log.msg(Log.LOG_DEBUG, 'GPU zone fan controller enabled')
            self.gpu_zone = GpuZone(self.log, self.ipmi, self.config)

        # Calculate the default sleep time for the main loop.
        polling_set = set()
        if self.cpu_zone_enabled:
            polling_set.add(self.cpu_zone.polling)
        if self.hd_zone_enabled:
            polling_set.add(self.hd_zone.polling)
        if self.gpu_zone_enabled:
            polling_set.add(self.gpu_zone.polling)

        # If none of the fan controller zones is enabled.
        if len(polling_set) == 0:
            self.log.msg(Log.LOG_ERROR, 'None of the zones are enabled, service terminated.')
            sys.exit(10)

        # Calculate the wait time in the main loop.
        wait = min(polling_set) / 2
        self.log.msg(Log.LOG_DEBUG, f'Main loop sleep time = {wait} sec')

        # Main execution loop.
        while True:
            if self.cpu_zone_enabled:
                self.cpu_zone.run()
            if self.hd_zone_enabled:
                self.hd_zone.run()
            if self.gpu_zone_enabled:
                self.gpu_zone.run()
            time.sleep(wait)


# End.



#
#    __init__.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#
from smfc.log import Log
from smfc.ipmi import Ipmi
from smfc.fancontroller import FanController
from smfc.cpuzone import CpuZone
from smfc.hdzone import HdZone
from smfc.gpuzone import GpuZone
from smfc.service import Service
from smfc.cmd import main

__all__ = [ "Log", "Ipmi", "FanController", "CpuZone", "HdZone", "GpuZone", "Service", "main"]

# End.



#
#   fancontroller.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.FanController() class implementation.
#
import os
import time
import re
from typing import List, Callable
from pyudev import Context, Device
from smfc.ipmi import Ipmi
from smfc.log import Log


class FanController:
    """Generic fan controller class for an IPMI zone."""

    # Constant values for temperature calculation
    CALC_MIN: int = 0
    CALC_AVG: int = 1
    CALC_MAX: int = 2

    # Configuration parameters
    log: Log                # Reference to a Log class instance
    ipmi: Ipmi              # Reference to an Ipmi class instance
    ipmi_zone: List[int]    # List of IPMI zones assigned to this fan controller
    name: str               # Name of the controller
    count: int              # Number of controlled entities
    temp_calc: int          # Calculate of the temperature (0-min, 1-avg, 2-max)
    steps: int              # Discrete steps in temperatures and fan levels
    sensitivity: float      # Temperature change to activate fan controller (C)
    polling: float          # Polling interval to read temperature (sec)
    min_temp: float         # Minimum temperature value (C)
    max_temp: float         # Maximum temperature value (C)
    min_level: int          # Minimum fan level (0..100%)
    max_level: int          # Maximum fan level (0..100%)
    hwmon_path: List[str]   # List of paths for HWMON devices

    # Measured or calculated attributes
    temp_step: float        # A temperature steps value (C)
    level_step: float       # A fan level step value (0..100%)
    last_time: float        # Last system time we polled temperature (timestamp)
    last_temp: float        # Last measured temperature value (C)
    last_level: int         # Last configured fan level (0..100%)

    # Function variable for selected temperature calculation method
    get_temp_func: Callable[[], float]

    def __init__(self, log: Log, ipmi: Ipmi, ipmi_zone: str, name: str, count: int, temp_calc: int,
                 steps: int, sensitivity: float, polling: float, min_temp: float, max_temp: float, min_level: int,
                 max_level: int) -> None:
        """Initialize the FanController class. Will raise an exception in case of invalid parameters.
        Args:
            log (Log): reference to a Log class instance
            ipmi (Ipmi): reference to an Ipmi class instance
            ipmi_zone (str): IPMI zone(s) assigned to the controller
            name (str): name of the controller
            count (int): number of devices
            temp_calc (int): calculation of temperature
            steps (int): discrete steps in temperatures and fan levels
            sensitivity (float): temperature change to activate fan controller (C)
            polling (float): polling time interval for reading temperature (sec)
            min_temp (float): minimum temperature value (C)
            max_temp (float): maximum temperature value (C)
            min_level (int): minimum fan level value [0..100%]
            max_level (int): maximum fan level value [0..100%]
        Raises:
            ValueError: invalid input parameter
        """
        # Save and validate configuration parameters.
        self.log = log
        self.ipmi = ipmi
        # Read the list of IPMI zones from a string (trim and remove multiple spaces, convert strings to integers)
        zone_str = re.sub(' +', ' ', ipmi_zone.strip())
        try:
            self.ipmi_zone = [int(s) for s in zone_str.split(',' if ',' in ipmi_zone else ' ')]
        except ValueError as e:
            raise e
        for zone in self.ipmi_zone:
            if zone not in range(0, 101):
                raise ValueError(f'invalid value: ipmi_zone={ipmi_zone}.')
        self.name = name
        self.count = count
        if self.count <= 0:
            raise ValueError('invalid value: count <= 0')
        self.temp_calc = temp_calc
        if self.temp_calc not in {self.CALC_MIN, self.CALC_AVG, self.CALC_MAX}:
            raise ValueError(f'invalid value: temp_calc ({temp_calc}).')
        self.steps = steps
        if self.steps <= 0:
            raise ValueError('invalid value: steps <= 0')
        self.sensitivity = sensitivity
        if self.sensitivity <= 0:
            raise ValueError('invalid value: sensitivity <= 0')
        self.polling = polling
        if self.polling < 0:
            raise ValueError('polling < 0')
        if max_temp < min_temp:
            raise ValueError('invalid value: max_temp < min_temp')
        self.min_temp = min_temp
        self.max_temp = max_temp
        if max_level < min_level:
            raise ValueError('invalid value: max_level < min_level')
        self.min_level = min_level
        self.max_level = max_level

        # Set the proper temperature function.
        if self.count == 1:
            self.get_temp_func = self.get_1_temp
        else:
            self.get_temp_func = self.get_avg_temp
            if self.temp_calc == self.CALC_MIN:
                self.get_temp_func = self.get_min_temp
            elif self.temp_calc == self.CALC_MAX:
                self.get_temp_func = self.get_max_temp

        # Try to read device temperature (the hwmon_path[] list has already been created by a child class).
        # If there is any problem with reading temperature, the program will stop here with an exception.
        self.get_temp_func()

        # Initialize calculated values.
        self.temp_step = (max_temp - min_temp) / steps
        self.level_step = (max_level - min_level) / steps
        self.last_temp = 0
        self.last_level = 0
        self.last_time = time.monotonic() - (polling + 1)
        # Print configuration at DEBUG log level.
        if self.log.log_level >= Log.LOG_CONFIG:
            self.log.msg(Log.LOG_CONFIG, f'{self.name} fan controller was initialized with:')
            self.log.msg(Log.LOG_CONFIG, f'   ipmi zone = {self.ipmi_zone}')
            self.log.msg(Log.LOG_CONFIG, f'   count = {self.count}')
            self.log.msg(Log.LOG_CONFIG, f'   temp_calc = {self.temp_calc}')
            self.log.msg(Log.LOG_CONFIG, f'   steps = {self.steps}')
            self.log.msg(Log.LOG_CONFIG, f'   sensitivity = {self.sensitivity}')
            self.log.msg(Log.LOG_CONFIG, f'   polling = {self.polling}')
            self.log.msg(Log.LOG_CONFIG, f'   min_temp = {self.min_temp}')
            self.log.msg(Log.LOG_CONFIG, f'   max_temp = {self.max_temp}')
            self.log.msg(Log.LOG_CONFIG, f'   min_level = {self.min_level}')
            self.log.msg(Log.LOG_CONFIG, f'   max_level = {self.max_level}')
            if hasattr(self, 'hwmon_path'):
                self.log.msg(Log.LOG_CONFIG, f'   hwmon_path = {[p if p else "smartctl" for p in self.hwmon_path]}')
            self.print_temp_level_mapping()

    @staticmethod
    def get_hwmon_path(udevc: Context, parent_dev: Device) -> str:
        """A helper function to get HWMON path of a given parent device's associated hwmon

        Args:
            udevc (Context): pyudev Context
            parent_dev (Device): parent device
        Returns:
            str: path for a HWMON device
        """
        try:
            [hwmon_device] = udevc.list_devices(subsystem='hwmon', parent=parent_dev)
        except ValueError:
            # If parent_dev has zero (or more?) hwmon device in its subtree
            hwmon_device = None
        return os.path.join(hwmon_device.sys_path, 'temp1_input') if hwmon_device is not None else ''

    def _get_nth_temp(self, index: int) -> float:
        """Get the temperature of the 'nth' element in the hwmon list. This is an empty implementation."""

    def get_1_temp(self) -> float:
        """Get a single temperature of a controlled entity in the IPMI zone.

        Returns:
            float: single temperature of a controlled entity (C)
        """
        return self._get_nth_temp(0)

    def get_min_temp(self) -> float:
        """Get the minimum temperature of multiple controlled entities.

        Returns:
            float: minimum temperature of the controlled entities (C)
        """
        minimum: float      # Minimum temperature value

        # Calculate minimum temperature.
        minimum = 1000.0
        for i in range(self.count):
            minimum = min(self._get_nth_temp(i), minimum)
        return minimum

    def get_avg_temp(self):
        """Get the average temperature of the controlled entities in the IPMI zone.

           Returns:
                float: average temperature of the controlled entities (C)
        """
        average: float      # Average temperature
        counter: int        # Value counter

        # Calculate average temperature.
        average = 0.0
        counter = 0
        for i in range(self.count):
            average += self._get_nth_temp(i)
            counter += 1
        return average / counter

    def get_max_temp(self) -> float:
        """Get the maximum temperature of the controlled entities in the IPMI zone.

           Returns:
                float: maximum temperature of the controlled entities (C)
        """
        maximum: float      # Maximum temperature value

        # Calculate minimum temperature.
        maximum = -1.0
        for i in range(self.count):
            maximum = max(self._get_nth_temp(i), maximum)
        return maximum

    def set_fan_level(self, level: int) -> None:
        """Set the new fan level in all IPMI zones of the controller.

        Args:
            level (int): new fan level [0..100]
        """
        self.ipmi.set_multiple_fan_levels(self.ipmi_zone, level)

    def callback_func(self) -> None:
        """Call-back function for a child class."""

    def run(self) -> None:
        """Run IPMI zone controller function with the following steps:

            * Step 1: Read current time. If the elapsed time is bigger than the polling time period
              then go to step 2, otherwise return.
            * Step 2: Read the current temperature. If the change of the temperature goes beyond
              the sensitivity limit then go to step 3, otherwise return
            * Step 3: Calculate the current gain and fan level based on the measured temperature
            * Step 4: If the new fan level is different it will be set and logged
        """
        current_time: float     # Current system timestamp (measured)
        current_temp: float     # Current temperature (measured)
        current_level: int      # Current fan level (calculated)
        current_gain: int       # Current gain (calculated)

        # Step 1: check the elapsed time.
        current_time = time.monotonic()
        if (current_time - self.last_time) >= self.polling:
            self.last_time = current_time

            # Step 2: read the temperature and check the sensitivity gap.
            self.callback_func()
            current_temp = self.get_temp_func()
            self.log.msg(Log.LOG_DEBUG, f'{self.name}: new temperature > {current_temp:.1f}C')
            if abs(current_temp - self.last_temp) >= self.sensitivity:
                self.last_temp = current_temp

                # Step 3: calculate gain and fan level.
                if current_temp <= self.min_temp:
                    current_level = self.min_level
                elif current_temp >= self.max_temp:
                    current_level = self.max_level
                else:
                    current_gain = int(round((current_temp - self.min_temp) / self.temp_step))
                    current_level = int(round(float(current_gain) * self.level_step)) + self.min_level

                # Step 4: the new fan level will be set and logged.
                if current_level != self.last_level:
                    self.last_level = current_level
                    self.set_fan_level(current_level)
                    self.log.msg(Log.LOG_INFO, f'{self.name}: new fan level > {current_level}%/{current_temp:.1f}C'
                                 f' @ IPMI {self.ipmi_zone} zone(s).')

    def print_temp_level_mapping(self) -> None:
        """Print out the user-defined temperature to level mapping value in log DEBUG level."""
        self.log.msg(Log.LOG_CONFIG, '   User-defined control function:')
        for i in range(self.steps + 1):
            self.log.msg(Log.LOG_CONFIG, f'   {i}. [T:{self.min_temp+(i*self.temp_step):.1f}C - '
                         f'L:{int(self.min_level + (i * self.level_step))}%]')

# End.



#
#   cpuzone.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.CpuZone() class implementation.
#
from configparser import ConfigParser
from pyudev import Context
from smfc.fancontroller import FanController
from smfc.ipmi import Ipmi
from smfc.log import Log


class CpuZone(FanController):
    """CPU zone fan control."""

    # Constant values for the configuration parameters.
    CS_CPU_ZONE: str = 'CPU zone'
    CV_CPU_ZONE_ENABLED: str = 'enabled'
    CV_CPU_IPMI_ZONE: str = 'ipmi_zone'
    CV_CPU_ZONE_TEMP_CALC: str = 'temp_calc'
    CV_CPU_ZONE_STEPS: str = 'steps'
    CV_CPU_ZONE_SENSITIVITY: str = 'sensitivity'
    CV_CPU_ZONE_POLLING: str = 'polling'
    CV_CPU_ZONE_MIN_TEMP: str = 'min_temp'
    CV_CPU_ZONE_MAX_TEMP: str = 'max_temp'
    CV_CPU_ZONE_MIN_LEVEL: str = 'min_level'
    CV_CPU_ZONE_MAX_LEVEL: str = 'max_level'

    def __init__(self, log: Log, udevc: Context, ipmi: Ipmi, config:ConfigParser) -> None:
        """Initialize the CpuZone class and raise exception in case of invalid configuration.
        Args:
            log (Log): reference to a Log class instance
            udevc (Context): reference to an udev database connection (instance of Context from pyudev)
            ipmi (Ipmi): reference to an Ipmi class instance
            config (ConfigParser): reference to the configuration (default=None)
        Raises:
            ValueError: multiple hwmon devices reported, one expected
            RuntimeError: No HWMON device found for CPU(s)
        """
        count: int  # CPU count.

        # Build the list of paths for hwmon devices.
        self.hwmon_path = []
        # We are looking for either Intel (coretemp) or AMD (k10temp) CPUs.
        for dev_filter in [{'MODALIAS':'platform:coretemp'}, {'DRIVER':'k10temp'}]:
            self.hwmon_path = [self.get_hwmon_path(udevc, dev) for dev in udevc.list_devices(**dev_filter)]
            # If we found results.
            if self.hwmon_path:
                break
        if not self.hwmon_path:
            raise RuntimeError('pyudev: No HWMON device(s) can be found for the CPU.')
        # Calculate count.
        count = len(self.hwmon_path)

        # Initialize FanController class.
        super().__init__(log, ipmi,
            config[CpuZone.CS_CPU_ZONE].get(CpuZone.CV_CPU_IPMI_ZONE, fallback=f'{Ipmi.CPU_ZONE}'),
            CpuZone.CS_CPU_ZONE, count,
            config[CpuZone.CS_CPU_ZONE].getint(CpuZone.CV_CPU_ZONE_TEMP_CALC, fallback=FanController.CALC_AVG),
            config[CpuZone.CS_CPU_ZONE].getint(CpuZone.CV_CPU_ZONE_STEPS, fallback=6),
            config[CpuZone.CS_CPU_ZONE].getfloat(CpuZone.CV_CPU_ZONE_SENSITIVITY, fallback=3.0),
            config[CpuZone.CS_CPU_ZONE].getfloat(CpuZone.CV_CPU_ZONE_POLLING, fallback=2),
            config[CpuZone.CS_CPU_ZONE].getfloat(CpuZone.CV_CPU_ZONE_MIN_TEMP, fallback=30.0),
            config[CpuZone.CS_CPU_ZONE].getfloat(CpuZone.CV_CPU_ZONE_MAX_TEMP, fallback=60.0),
            config[CpuZone.CS_CPU_ZONE].getint(CpuZone.CV_CPU_ZONE_MIN_LEVEL, fallback=35),
            config[CpuZone.CS_CPU_ZONE].getint(CpuZone.CV_CPU_ZONE_MAX_LEVEL, fallback=100)
        )

    def _get_nth_temp(self, index: int) -> float:
        """Get the temperature of the 'nth' element in the hwmon list.
           Args:
               index (int): index in hwmon list
           Returns:
               float: temperature value
           Raises:
               FileNotFoundError:   file not found
               IOError:             file cannot be opened
               ValueError:          invalid value read from file
               IndexError:          invalid index
           """
        value: float  # Temperature value

        try:
            with open(self.hwmon_path[index], "r", encoding="UTF-8") as f:
                value = float(f.read()) / 1000
        except (IOError, FileNotFoundError, ValueError) as e:
            raise e
        return value

# End.



#!/usr/bin/env python3
#
#   test_03_fancontroller.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.FanController() class.
#
import time
import re
from typing import List, Tuple
import pytest
import pyudev
from mock import MagicMock, call
from pytest_mock import MockerFixture
from smfc import FanController, Log, Ipmi, CpuZone, HdZone
from .test_00_data import MockDevice, MockContext

class TestFanController:
    """Unit test class for smfc.FanController() class"""

    #pylint: disable=line-too-long
    @pytest.mark.parametrize("ipmi_zone, name, count, temp_calc, steps, sensitivity, polling, min_temp, max_temp, "
                             "min_level, max_level, error", [
        ('0', CpuZone.CS_CPU_ZONE, 1, FanController.CALC_MIN, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 1'),
        ('0', CpuZone.CS_CPU_ZONE, 4, FanController.CALC_MIN, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 2'),
        ('0', CpuZone.CS_CPU_ZONE, 6, FanController.CALC_AVG, 6, 5, 4, 32, 52, 37, 95,  'FanController.__init__() 3'),
        ('0', CpuZone.CS_CPU_ZONE, 8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90,  'FanController.__init__() 4'),
        ('1', HdZone.CS_HD_ZONE,   1, FanController.CALC_MIN, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 5'),
        ('1', HdZone.CS_HD_ZONE,   4, FanController.CALC_MIN, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 6'),
        ('1', HdZone.CS_HD_ZONE,   6, FanController.CALC_AVG, 6, 5, 4, 32, 52, 37, 95,  'FanController.__init__() 7'),
        ('1', HdZone.CS_HD_ZONE,   8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90,  'FanController.__init__() 8'),
        ('0, 1',    HdZone.CS_HD_ZONE, 8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90,      'FanController.__init__() 9'),
        ('0, 1, 2', HdZone.CS_HD_ZONE, 8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90,      'FanController.__init__() 10'),
        (' 0 1 2',   HdZone.CS_HD_ZONE, 8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90,     'FanController.__init__() 11'),
        ('  0  1  2 ', HdZone.CS_HD_ZONE, 8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90,   'FanController.__init__() 12'),
        ('  0,  1,  2 ', HdZone.CS_HD_ZONE, 8, FanController.CALC_MAX, 7, 6, 6, 34, 54, 39, 90, 'FanController.__init__() 13')
    ])
    #pylint: enable=line-too-long
    def test_init_p1(self, mocker:MockerFixture, ipmi_zone: str, name: str, count: int, temp_calc: int,
                     steps: int, sensitivity: float, polling: float, min_temp: float, max_temp: float, min_level: int,
                     max_level, error: str) -> None:
        """Positive unit test for FanController.__init__() method. It contains the following steps:
            - mock print(), FanController._get_nth_temp() functions
            - initialize a Log, Ipmi, and FanController classes
            - ASSERT: if the class attributes contain different values that were passed to __init__
            - delete all instances
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_get_nth_temp = MagicMock()
        mocker.patch('smfc.FanController._get_nth_temp', mock_get_nth_temp)
        mock_get_nth_temp.return_value = 38.5
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_fc = FanController(my_log, my_ipmi, ipmi_zone, name, count, temp_calc, steps, sensitivity, polling,
                              min_temp, max_temp, min_level, max_level)
        assert my_fc.log == my_log, error
        assert my_fc.ipmi == my_ipmi
        zone_str = re.sub(' +', ' ', ipmi_zone.strip())
        assert my_fc.ipmi_zone == [int(s) for s in zone_str.split(',' if ',' in ipmi_zone else ' ')], error
        assert my_fc.name == name, error
        assert my_fc.count == count, error
        assert my_fc.temp_calc == temp_calc, error
        assert my_fc.steps == steps, error
        assert my_fc.sensitivity == sensitivity, error
        assert my_fc.polling == polling, error
        assert my_fc.min_temp == min_temp, error
        assert my_fc.max_temp == max_temp, error
        assert my_fc.min_level == min_level, error
        assert my_fc.max_level == max_level, error
        assert my_fc.level_step == (max_level - min_level) / steps, error
        assert my_fc.last_temp == 0, error
        assert my_fc.last_level == 0, error

    @pytest.mark.parametrize("ipmi_zone, name, count, temp_calc, steps, sensitivity, polling, min_temp, max_temp, "
                             "min_level, max_level, error", [
        # ipmi_zone is invalid
        ('-1',      CpuZone.CS_CPU_ZONE, 1, 0, 5, 4, 2, 30, 50, 35, 100,   'FanController.__init__() 14'),
        ('101',     CpuZone.CS_CPU_ZONE, 1, 0, 5, 4, 2, 30, 50, 35, 100,   'FanController.__init__() 15'),
        ('%, &',    CpuZone.CS_CPU_ZONE, 1, 0, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 16'),
        ('1; 2; 3', CpuZone.CS_CPU_ZONE, 1, 0, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 17'),
        ('1, %, 3', CpuZone.CS_CPU_ZONE, 1, 0, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 18'),
        # count <= 0
        ('0',   CpuZone.CS_CPU_ZONE, -1, 0, 5, 4, 2, 30, 50, 35, 100,  'FanController.__init__() 19'),
        ('0',   CpuZone.CS_CPU_ZONE, 0, 0, 5, 4, 2, 30, 50, 35, 100,   'FanController.__init__() 20'),
        # temp_calc is invalid
        ('0',   CpuZone.CS_CPU_ZONE, 1, -1, 5, 4, 2, 30, 50, 35, 100,  'FanController.__init__() 21'),
        ('0',   CpuZone.CS_CPU_ZONE, 1, 100, 5, 4, 2, 30, 50, 35, 100, 'FanController.__init__() 22'),
        # step <= 0
        ('1',   HdZone.CS_HD_ZONE,   1, 1, -2, 4, 2, 30, 50, 35, 100,  'FanController.__init__() 23'),
        ('1',   HdZone.CS_HD_ZONE,   1, 1, 0, 4, 2, 30, 50, 35, 100,   'FanController.__init__() 24'),
        # sensitivity <= 0
        ('1',   HdZone.CS_HD_ZONE,   1, 1, 5, 0, 2, 30, 50, 35, 100,   'FanController.__init__() 25'),
        ('1',   HdZone.CS_HD_ZONE,   1, 1, 5, -2, 2, 30, 50, 35, 100,  'FanController.__init__() 26'),
        # polling < 0
        ('1',   HdZone.CS_HD_ZONE,   1, 1, 5, 4, -2, 30, 50, 35, 100,  'FanController.__init__() 27'),
        # max_temp < min_temp
        ('1',   HdZone.CS_HD_ZONE,   1, 1, 5, 4, 2, 50, 30, 35, 100,   'FanController.__init__() 28'),
        # max_level < min_level
        ('1',   HdZone.CS_HD_ZONE,   1, 1, 5, 4, 2, 30, 50, 100, 35,   'FanController.__init__() 29')
    ])
    def test_init_n1(self, mocker: MockerFixture, ipmi_zone: str, name: str, count: int, temp_calc: int,
                     steps: int, sensitivity: float, polling: float, min_temp: float, max_temp: float,
                     min_level: int, max_level, error: str) -> None:
        """Negative unit test for FanController.__init__() method. It contains the following steps:
            - mock print(), FanController._get_nth_temp() functions
            - initialize a Config, Log, Ipmi, and FanController classes
            - ASSERT: if exception was not raised in case of invalid parameter values
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_get_nth_temp = MagicMock()
        mocker.patch('smfc.FanController._get_nth_temp', mock_get_nth_temp)
        mock_get_nth_temp.return_value = 38.5
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        with pytest.raises(ValueError) as cm:
            FanController(my_log, my_ipmi, ipmi_zone, name, count, temp_calc, steps, sensitivity, polling,
                          min_temp, max_temp, min_level, max_level)
        assert cm.type is ValueError, error

    @pytest.mark.parametrize("devices, result, error", [
        # Normal case - 1 device found
        (['/sys'],              '/sys/temp1_input', 'FanController.get_hwmon_path() 1'),
        # Error case - multiple devices found
        (['/sys1', '/sys2'],    '',                 'FanController.get_hwmon_path() 2'),
        # Error case - no devices found
        ([],                    '',                 'FanController.get_hwmon_path() 3')
    ])
    def test_get_hwmon_path(self, mocker:MockerFixture, devices: List[str], result:str, error: str) -> None:
        """Positive unit test for FanController.get_hwmon_path() method. It contains the following steps:
            - mock pyudev.Context and pyudev.Device classes
            - initialize a pyudev.Context() and a pyudev.Device() classes
            - call FanController.get_hwmon_path() function
            - ASSERT: if not the expected result returned
        """
        result_devices: List[MockDevice]
        parent: pyudev.Device
        context: pyudev.Context

        # Result Device list for pyudev.list_devices()
        result_devices = []
        if devices is not None:
            for i in devices:
                md = MockDevice()
                #pylint: disable=protected-access
                md._sys_path = i
                # pylint: enable=protected-access
                result_devices.append(md)
        mocker.patch.object(pyudev.Context, '__new__', return_value=MockContext(result_devices))
        mocker.patch.object(pyudev.Device, '__new__', return_value=MockDevice())
        context = pyudev.Context.__new__(pyudev.Context)
        parent = pyudev.Device.__new__(pyudev.Device)
        assert FanController.get_hwmon_path(context, parent) == result, error

    @pytest.mark.parametrize("count, code, temps, expected, error", [
        # get_1_temp()
        (1, 1, [38.5], 38.5, 'fc get_1_temp 1'),
        # get_min_temp()
        (3, 2, [38.5, 38.5, 38.5], 38.5, 'fc get_min_temp 1'),
        (3, 2, [38.5, 40.5, 42.5], 38.5, 'fc get_min_temp 2'),
        # get_avg_temp()
        (3, 3, [38.5, 38.5, 38.5], 38.5, 'fc get_avg_temp 1'),
        (3, 3, [38.5, 40.5, 42.5], 40.5, 'fc get_avg_temp 2'),
        (8, 3, [38.0, 40.0, 42.0, 44.0, 46.0, 48.0, 50.0, 52.0], 45.0, 'fc get_avg_temp 3'),
        # get_max_temp()
        (3, 4, [38.5, 38.5, 38.5], 38.5, 'fc get_max_temp 1'),
        (3, 4, [38.5, 40.5, 42.5], 42.5, 'fc get_max_temp 2'),
        (8, 4, [38.0, 40.0, 42.0, 44.0, 46.0, 48.0, 50.0, 52.0], 52.0, 'fc get_max_temp 3')
    ])
    def test_get_xxx_temp(self, mocker: MockerFixture, count: int, code: int, temps: List[float], expected: float,
                          error: str):
        """Primitive positive test function. It contains the following steps:
            - mock FanController._get_nth_temp() function
            - initialize an empty FanController class
            - ASSERT: if get_xxx_temp() functions return different from expected temperature
        """
        t: float    # temperature

        my_fc = FanController.__new__(FanController)
        my_fc.count = count
        mock_temp = MagicMock()
        mock_temp.side_effect = temps
        mocker.patch('smfc.FanController._get_nth_temp', mock_temp)
        if code == 1:
            t = my_fc.get_1_temp()
        elif code == 2:
            t = my_fc.get_min_temp()
        elif code == 3:
            t = my_fc.get_avg_temp()
        else:  # code == 4:
            t = my_fc.get_max_temp()
        assert t == expected, error

    @pytest.mark.parametrize("zones, level, error", [
        ([0],       45, 'FanController.set_fan_level() 1'),
        ([1],       55, 'FanController.set_fan_level() 2'),
        ([0, 1],    65, 'FanController.set_fan_level() 3'),
        ([0, 1, 2], 75, 'FanController.set_fan_level() 4')
    ])
    def test_set_fan_level(self, mocker: MockerFixture, zones: List[int], level: int, error: str):
        """Positive unit test for FanController.set_fan_level() method. It contains the following steps:
            - mock Ipmi.set_fan_level() functions
            - initialize an empty FanController class
            - ASSERT: if the Ipmi.set_fan_level() function was called with different parameters
        """
        my_ipmi = Ipmi.__new__(Ipmi)
        my_fc = FanController.__new__(FanController)
        my_fc.ipmi_zone = zones
        my_fc.ipmi = my_ipmi
        mock_set_multiple_fan_levels = MagicMock()
        mocker.patch('smfc.Ipmi.set_multiple_fan_levels', mock_set_multiple_fan_levels)
        my_fc.set_fan_level(level)
        calls = []
        for z in my_fc.ipmi_zone:
            calls.append(call(z, level))
        mock_set_multiple_fan_levels.assert_called_with(zones, level)
        assert mock_set_multiple_fan_levels.call_count == 1, error

    @pytest.mark.parametrize(
        "steps, sensitivity, polling, min_temp, max_temp, min_level, max_level, temp, level, error", [
        (5, 1, 1, 30, 50, 35, 100, None, None, 'FanController.run() 1'),
        (5, 1, 1, 40, 40, 45, 45, None, None,  'FanController.run() 2'),
        # Check level if temperature is under the minimum value.
        (5, 1, 1, 30, 50, 35, 100, 25.0, 35,   'FanController.run() 3'),
        # Check level if temperature is above the maximum value.
        (5, 1, 1, 30, 50, 35, 100, 55.0, 100,  'FanController.run() 4')
        ]
    )
    def test_run(self, mocker: MockerFixture, steps: int, sensitivity: float, polling: float, min_temp: float,
                  max_temp: float, min_level: int, max_level, temp: float, level: int, error: str) -> None:
        """Primitive positive test function. It contains the following steps:
            - mock print() and FanController._get_nth_temp() functions
            - initialize an empty FanController class
            - ASSERT: if the run() generates different fan level based on the input zone temperature
            - delete the instances
        """

        # Test data set 1 for a generic configuration (dynamic mapping):
        # steps=5, min_temp=30, max_temp=50, min_level=35, max_level=100
        test_values_1: List[Tuple[float, int]] = [
            (30.0, 35), (31.0, 35), (32.0, 35), (33.0, 48), (34.0, 48), (35.0, 48), (36.0, 61), (37.0, 61),
            (38.0, 61), (39.0, 61), (40.0, 61), (41.0, 74), (42.0, 74), (43.0, 74), (44.0, 87), (45.0, 87),
            (46.0, 87), (47.0, 87), (48.0, 87), (49.0, 100), (50.0, 100)
            ]
        # Test data set 2 for special configuration (constant mapping):
        # steps=5, min_temp=40, max_temp=40, min_level=45, max_level=45
        test_values_2: List[Tuple[float, int]] = [
            (30.0, 45), (31.0, 45), (32.0, 45), (33.0, 45), (34.0, 45), (35.0, 45), (36.0, 45), (37.0, 45),
            (38.0, 45), (39.0, 45), (40.0, 45), (41.0, 45), (42.0, 45), (43.0, 45), (44.0, 45), (45.0, 45),
            (46.0, 45), (47.0, 45), (48.0, 45), (49.0, 45), (50.0, 45)
            ]

        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_ipmi_set_fan_level = MagicMock()
        mocker.patch('smfc.FanController.set_fan_level', mock_ipmi_set_fan_level)
        mock_temp = MagicMock()
        mock_temp.return_value = 0
        mocker.patch('smfc.FanController._get_nth_temp', mock_temp)
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)

        my_fc = FanController(my_log, my_ipmi, '0', CpuZone.CS_CPU_ZONE, 1, FanController.CALC_AVG,
                              steps, sensitivity, polling, min_temp, max_temp, min_level, max_level)

        # If temperature/level is not specified, we use the internal data sets.
        if temp is None:
            # Test 1 with a valid data set.
            if min_temp < max_temp:
                for i in test_values_1:
                    mock_temp.return_value=i[0]
                    my_fc.last_level = 0
                    my_fc.last_time = time.monotonic() - (polling + 1)
                    my_fc.run()
                    assert my_fc.last_temp == i[0], error
                    assert my_fc.last_level == i[1], error
                    if mock_ipmi_set_fan_level.call_count > 0:
                        mock_ipmi_set_fan_level.assert_called_with(i[1])
            # Test 2 with constant mapping.
            elif min_temp == max_temp:
                for i in test_values_2:
                    mock_temp.return_value=i[0]
                    my_fc.last_level = 0
                    my_fc.last_time = time.monotonic() - (polling + 1)
                    my_fc.run()
                    assert my_fc.last_temp == i[0], error
                    assert my_fc.last_level == i[1], error
                    if mock_ipmi_set_fan_level.call_count > 0:
                        mock_ipmi_set_fan_level.assert_called_with(i[1])
        # Test 3 - special cases with specific temp/level values.
        else:
            mock_temp.return_value = temp
            my_fc.last_level = 0
            my_fc.last_time = time.monotonic() - (polling + 1)
            my_fc.run()
            assert my_fc.last_temp == temp, error
            assert my_fc.last_level == level, error
            if mock_ipmi_set_fan_level.call_count > 0:
                mock_ipmi_set_fan_level.assert_called_with(level)


# End.



#!/usr/bin/env python3
#
#   test_00_data.py (C) 2022-2025, Peter Sulyok
#   Test data handling class for unit tests.
#
import configparser
import os
import random
import shutil
import tempfile
from typing import List
from pyudev import DeviceNotFoundByFileError


class TestData:
    """Class for test data handling."""

    td_dir: str = ''                    # Test data directory
    cpu_files: List[str] = []           # CPU hwmon files

    hd_names: str                       # HD names in configuration parameter form
    hd_name_list: List[str] = []        # HD names in a list
    hd_files: List[str] = []            # HD hwmon files

    def __init__(self):
        """Initialize the class. It creates a temporary directory."""
        self.td_dir = tempfile.mkdtemp()

    def __del__(self):
        """It deletes the temporary directory with is all content."""
        shutil.rmtree(self.td_dir)

    def create_cpu_data(self, count: int, temp_list: List[float] = None) -> None:
        """Generic method to create temporary test data files (similarly to hwmon naming convention and content)."""
        hwmon_file: str

        self.cpu_files = []
        for i in range(count):
            hwmon_file = os.path.join(self.td_dir, 'cpu', 'coretemp.' + str(i), 'hwmon')
            os.makedirs(hwmon_file, exist_ok=True)
            hwmon_file = os.path.join(hwmon_file, 'temp1_input')
            with open(hwmon_file, "w+t", encoding="UTF-8") as f:
                if temp_list:
                    v = temp_list[i]
                else:
                    v = random.uniform(30.0, 60.0)
                f.write(str(v * 1000))
            self.cpu_files.append(hwmon_file)

    def create_hd_data(self, count: int, temp_list: List[float] = None) -> None:
        """Generic method to create temporary test data files (similarly to hwmon naming convention and content)."""
        letters: List[str] = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q' ]
        hwmon_path: str
        disk_name: str

        self.hd_names = ''
        self.hd_name_list = []
        self.hd_files = []
        separator = random.choice([' ', '\n'])
        for i in range(count):
            disk_name = '/dev/sd' + letters[i]
            self.hd_names += disk_name + separator
            self.hd_name_list.append(disk_name)
            hwmon_path = os.path.join(self.td_dir, 'disks', str(i)+':0:0:0', 'hwmon')
            os.makedirs(hwmon_path, exist_ok=True)
            hwmon_path = os.path.join(hwmon_path, 'temp1_input')
            with open(hwmon_path, "w+t", encoding="UTF-8") as f:
                if temp_list:
                    v = temp_list[i]
                else:
                    v = random.uniform(32.0, 45.0)
                v *= 1000
                f.write(f"{v:.0f}")
            self.hd_files.append(hwmon_path)

    def create_config_file(self, my_config: configparser.ConfigParser) -> str:
        """Creates a config file. """
        h, name = tempfile.mkstemp(prefix='config', suffix='.conf', dir=self.td_dir)
        with os.fdopen(h, "w+t") as f:
            my_config.write(f)
        return name

    def create_command_file(self, content: str = 'echo OK') -> str:
        """Creates an executable bash script. """
        h, name = tempfile.mkstemp(suffix='.sh', dir=self.td_dir)
        with os.fdopen(h, "w+t") as f:
            f.write(str('#!/bin/bash\n'))
            f.write(str(content + '\n'))
        os.system('chmod +x ' + name)
        return name

    @staticmethod
    def delete_file(path: str) -> None:
        """Deletes the specified file."""
        os.remove(path)

    def create_ipmi_command(self) -> str:
        """Creates a bash script emulating ipmitool."""
        return self.create_command_file("""
# ipmitool emulation

if [[ $1 = "sdr" ]] ; then
	echo OK
fi

# IPMI get fan mode (raw 0x30 0x45 0x00)
if [[ $1 = "raw" && $2 = "0x30" && $3 = "0x45" && $4 = "0x00" ]] ; then
  r=$((1 + (RANDOM % 4)))
  if [[ "$r" -eq "3" ]] ; then
    r=1
  fi
	echo "$r"
	exit 0
fi

# IPMI set fan mode (raw 0x30 0x45 0x01)
if [[ $1 = "raw" && $2 = "0x30" && $3 = "0x45" && $4 = "0x01" ]] ; then
	exit 0
fi

# IPMI set fan level (raw 0x30 0x70 0x66 0x01)
if [[ $1 = "raw" && $2 = "0x30" && $3 = "0x70" && $4 = "0x66" && $5 = "0x01" ]] ; then
	exit 0
fi
        """)

    def create_smart_command(self) -> str:
        """Creates a shell script emulating `smartctl`."""
        return self.create_command_file("""
# smartctl emulation script.

# Print header
cat << EOF
smartctl 7.3 2022-02-28 r5338 [x86_64-linux-6.1.0-32-amd64] (local build)
Copyright (C) 2002-22, Bruce Allen, Christian Franke, www.smartmontools.org

EOF

# smartctl -a /dev/sd?
if [[ $1 = "-a" ]] ; then
    r=$((RANDOM % 3))	
    case "$r" in 
        "0")
            echo "Current Drive Temperature:     37 C" ;;
        "1")
            echo "190 Airflow_Temperature_Cel 0x0032   075   045   000    Old_age   Always       -       25" ;;
        "2")
            echo "194 Temperature_Celsius     0x0002   232   232   000    Old_age   Always       -       28 (Min/Max 17/45)" ;;
    esac
fi

# smartctl -i -n standby /dev/sd?
if [[ $1 = "-i" && $2 = "-n" && $3 = "standby" ]] ; then
    r=$((RANDOM % 2))	
    case "$r" in 
        "0")
            cat << EOF
=== START OF INFORMATION SECTION ===
Model Family:     Samsung based SSDs
Device Model:     Samsung SSD 870 QVO 8TB
Serial Number:    S5SSNG1NB01829M
LU WWN Device Id: 5 002538 f70b0ee2f
Firmware Version: SVQ01B6Q
User Capacity:    8,001,563,222,016 bytes [8.00 TB]
Sector Size:      512 bytes logical/physical
Rotation Rate:    Solid State Device
Form Factor:      2.5 inches
TRIM Command:     Available, deterministic, zeroed
Device is:        In smartctl database [for details use: -P show]
ATA Version is:   ACS-4 T13/BSR INCITS 529 revision 5
SATA Version is:  SATA 3.3, 6.0 Gb/s (current: 6.0 Gb/s)
Local Time is:    Sat May 15 14:26:26 2021 CEST
SMART support is: Available - device has SMART capability.
SMART support is: Enabled
Power mode is:    ACTIVE or IDLE
EOF
            r=0 ;;
            
        "1")
            cat << EOF
Device is in STANDBY mode, exit(2)
EOF
            r=2 ;;            
    esac
    exit $r
fi

# smartctl -s standby,now /dev/sd?
if [[ $1 = "-s" && $2 = "standby,now" ]] ; then
    echo "Device placed in STANDBY mode"
    exit 0  
fi

exit 0
"""
        )

    def create_nvidia_smi_command(self, count: int, temp_list: List[float] = None) -> str:
        """Creates a shell script emulating `nvidia-smi`."""
        file_content = 'cat << EOF\n'
        for i in range(0, count):
            if temp_list:
                v = temp_list[i]
            else:
                v = random.uniform(35.0, 75.0)
            file_content += f'{v:.0f}' + '\n'
        file_content += 'EOF\n'
        return self.create_command_file(file_content)

    def create_text_file(self, content: str) -> str:
        """Creates a text file with the specified content."""
        h, name = tempfile.mkstemp(prefix='text', suffix='.txt', dir=self.td_dir)
        with os.fdopen(h, "w+t") as f:
            f.write(content)
        return name


#pylint: disable=missing-function-docstring
#pylint: disable=too-few-public-methods
class MockDevice:
    """Mock class for pyudev.Device() class"""

    _sys_path: str

    def __init__(self, context=None, _device=None):
        pass

    def __del__(self):
        pass

    @property
    def parent(self):
        return None

    @property
    def sys_path(self):
        return self._sys_path


def factory_mockdevice():
    """Can generate MockDevice() class."""
    return MockDevice()


class MockContext:
    """Mock class for pyudev.Context() class."""
    mocked_devices: List[MockDevice]

    def __init__(self, devices=None):
        self.mocked_devices = devices

    def __del__(self):
        pass

    #pylint: disable=unused-argument
    def list_devices(self, **kwargs):
        return iter(self.mocked_devices)
    # pylint: enable=unused-argument


class MockDevices:
    """Mock class for pyudev.Devices() class."""

    @classmethod
    def from_device_file(cls, context=None, filename=None):
        if filename == "raise":
            raise DeviceNotFoundByFileError()
        return MockDevice(context)


class MockedContextError:
    """Mock class for pyudev.Context() class will generate ImportError exception."""
    def __init__(self):
        raise ImportError


class MockedContextGood:
    """Mock class for pyudev.Context() class will not generate any exception."""
    def __init__(self):
        pass

#pylint: enable=missing-function-docstring
#pylint: enable=too-few-public-methods


# End.



#
#   ipmi.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.Ipmi() class implementation.
#
import subprocess
import time
from configparser import ConfigParser
from typing import List
from smfc.log import Log


class Ipmi:
    """IPMI interface class can set/get IPMI fan mode, and can set IPMI fan level using ipmitool."""

    log: Log                    # Reference to a Log class instance
    command: str                # Full path for ipmitool command.
    fan_mode_delay: float       # Delay time after execution of IPMI set fan mode function
    fan_level_delay: float      # Delay time after execution of IPMI set fan level function
    remote_parameters: str      # Remote IPMI parameters
    sudo: bool                  # Use `sudo` command for `ipmitool` command

    # Constant values for IPMI fan modes:
    STANDARD_MODE: int = 0
    FULL_MODE: int = 1
    OPTIMAL_MODE: int = 2
    PUE_MODE: int = 3
    HEAVY_IO_MODE: int = 4

    # Constant values for IPMI fan zones:
    CPU_ZONE: int = 0
    HD_ZONE: int = 1

    # Constant values for the results of IPMI operations:
    SUCCESS: int = 0
    ERROR: int = -1

    # Constant values for the configuration parameters.
    CS_IPMI: str = 'Ipmi'
    CV_IPMI_COMMAND: str = 'command'
    CV_IPMI_FAN_MODE_DELAY: str = 'fan_mode_delay'
    CV_IPMI_FAN_LEVEL_DELAY: str = 'fan_level_delay'
    CV_IPMI_REMOTE_PARAMETERS: str = 'remote_parameters'

    def __init__(self, log: Log, config: ConfigParser, sudo: bool) -> None:
        """Initialize the Ipmi class with a log class and with a configuration class.
        Args:
            log (Log): a Log class instance
            config (ConfigParser): a ConfigParser class instance
            sudo (bool): sudo flag
        Raises:
            ValueError: invalid input parameters
            FileNotFoundError: ipmitool not found
            RuntimeError: ipmitool execution error
        """
        # Set default or read from configuration
        self.log = log
        self.command = config[Ipmi.CS_IPMI].get(Ipmi.CV_IPMI_COMMAND, '/usr/bin/ipmitool')
        self.fan_mode_delay = config[Ipmi.CS_IPMI].getint(Ipmi.CV_IPMI_FAN_MODE_DELAY, fallback=10)
        self.fan_level_delay = config[Ipmi.CS_IPMI].getint(Ipmi.CV_IPMI_FAN_LEVEL_DELAY, fallback=2)
        self.remote_parameters = config[Ipmi.CS_IPMI].get(Ipmi.CV_IPMI_REMOTE_PARAMETERS, fallback='')
        self.sudo = sudo

        # Validate configuration
        # Check 1: a valid command can be executed successfully.
        try:
            self._exec_ipmitool(['sdr'])
        except (FileNotFoundError, RuntimeError) as e:
            raise e
        # Check 2: fan_mode_delay must be positive.
        if self.fan_mode_delay < 0:
            raise ValueError(f'Negative fan_mode_delay= parameter ({self.fan_mode_delay})')
        # Check 3: fan_mode_delay must be positive.
        if self.fan_level_delay < 0:
            raise ValueError(f'Negative fan_level_delay= parameter ({self.fan_level_delay})')
        # Print the configuration out at DEBUG log level.
        if self.log.log_level >= Log.LOG_CONFIG:
            self.log.msg(Log.LOG_CONFIG, 'Ipmi module was initialized with:')
            self.log.msg(Log.LOG_CONFIG, f'   {Ipmi.CV_IPMI_COMMAND} = {self.command}')
            self.log.msg(Log.LOG_CONFIG, f'   {Ipmi.CV_IPMI_FAN_MODE_DELAY} = {self.fan_mode_delay}')
            self.log.msg(Log.LOG_CONFIG, f'   {Ipmi.CV_IPMI_FAN_LEVEL_DELAY} = {self.fan_level_delay}')
            self.log.msg(Log.LOG_CONFIG, f'   {Ipmi.CV_IPMI_REMOTE_PARAMETERS} = {self.remote_parameters}')

    def _exec_ipmitool(self, args: List[str]) -> subprocess.CompletedProcess:
        """Execute `ipmitool` command.
        Args:
            args(List[str]): command line parameters
        Returns:
            subprocess.CompletedProcess: result of the executed subprocess
        Raises:
            FileNotFoundError: ipmitool cannot be found
            RuntimeError: ipmitool execution problem (e.g. non-root user, incompatible IPMI system/motherboard)
        """
        r: subprocess.CompletedProcess      # result of the executed process
        arguments: List[str]                # Command arguments

        try:
            # Construct command line parameters.
            arguments = []
            # Add `sudo` if needed.
            if self.sudo:
                arguments.append('sudo')
            # Add `ipmitool` path.
            arguments.append(self.command)
            # Add remote parameters if needed.
            if self.remote_parameters:
                arguments.extend(self.remote_parameters.split())
            # Add additional command line parameters from caller.
            arguments.extend(args)
            r = subprocess.run(arguments, check=False, capture_output=True, text=True)
            # Check error code.
            if r.returncode != 0:
                if self.sudo and 'sudo' in r.stderr:
                    raise RuntimeError(f'sudo error ({r.returncode}): {r.stderr}.')
                raise RuntimeError(f'ipmitool error ({r.returncode}): {r.stderr}.')
        except FileNotFoundError as e:
            raise e
        return r

    def get_fan_mode(self) -> int:
        """Get the current IPMI fan mode.
        Returns:
            int: fan mode (ERROR, STANDARD_MODE, FULL_MODE, OPTIMAL_MODE, HEAVY_IO_MODE)
        Raises:
            FileNotFoundError: ipmitool cannot be found
            RuntimeError: ipmitool execution problem (e.g. non-root user, incompatible IPMI system/motherboard)
            ValueError: output of the ipmitool cannot be interpreted/converted
        """
        r: subprocess.CompletedProcess  # result of the executed process
        m: int                          # fan mode

        # Read the current IPMI fan mode.
        try:
            r = self._exec_ipmitool(['raw', '0x30', '0x45', '0x00'])
            m = int(r.stdout)
        except (RuntimeError, FileNotFoundError) as e:
            raise e
        return m

    @staticmethod
    def get_fan_mode_name(mode: int) -> str:
        """Get the name of the specified IPMI fan mode.
        Args:
            mode (int): fan mode
        Returns:
            str: name of the fan mode ('UNKNOWN', 'STANDARD', 'FULL', 'OPTIMAL', 'PUE' 'HEAVY IO')
        """
        fan_mode_name: str  # Name of the fan mode

        fan_mode_name = 'UNKNOWN'
        if mode == Ipmi.STANDARD_MODE:
            fan_mode_name = 'STANDARD'
        elif mode == Ipmi.FULL_MODE:
            fan_mode_name = 'FULL'
        elif mode == Ipmi.OPTIMAL_MODE:
            fan_mode_name = 'OPTIMAL'
        elif mode == Ipmi.PUE_MODE:
            fan_mode_name = 'PUE'
        elif mode == Ipmi.HEAVY_IO_MODE:
            fan_mode_name = 'HEAVY IO'
        return fan_mode_name

    def set_fan_mode(self, mode: int) -> None:
        """Set the IPMI fan mode.
        Args:
            mode (int): fan mode (STANDARD_MODE, FULL_MODE, OPTIMAL_MODE, PUE_MODE, HEAVY_IO_MODE)
        Raises:
            ValueError: invalid input parameter
            FileNotFoundError: ipmitool command cannot be found
            RuntimeError: ipmitool execution problem (e.g. non-root user, incompatible IPMI system/motherboard)
        """
        # Validate mode parameter.
        if mode not in {self.STANDARD_MODE, self.FULL_MODE, self.OPTIMAL_MODE, self.PUE_MODE, self.HEAVY_IO_MODE}:
            raise ValueError(f'Invalid fan mode value ({mode}).')
        # Call ipmitool command and set the new IPMI fan mode.
        try:
            self._exec_ipmitool(['raw', '0x30', '0x45', '0x01', f'0x{mode:02x}'])
        except (RuntimeError, FileNotFoundError) as e:
            raise e
        # Give time for IPMI system/fans to apply changes in the new fan mode.
        time.sleep(self.fan_mode_delay)

    def set_fan_level(self, zone: int, level: int) -> None:
        """Set the fan level in the specified IPMI zone. Could raise several exceptions in case of invalid parameters.
        Args:
            zone (int): IPMI zone
            level (int): fan level in % (0-100)
        Raises:
            ValueError: invalid input parameter
            FileNotFoundError: ipmitool command cannot be found
            RuntimeError: ipmitool execution problem (e.g. non-root user, incompatible IPMI system/motherboard)
        """
        # Validate zone parameter
        if zone not in range(0, 101):
            raise ValueError(f'Invalid value: zone ({zone}).')
        # Validate level parameter (must be in the interval [0..100%])
        if level not in range(0, 101):
            raise ValueError(f'Invalid value: level ({level}).')
        # Set the new IPMI fan level in the specific zone
        try:
            self._exec_ipmitool(['raw', '0x30', '0x70', '0x66', '0x01', f'0x{zone:02x}', f'0x{level:02x}'])
        except (FileNotFoundError, RuntimeError) as e:
            raise e
        # Give time for IPMI and fans to spin up/down.
        time.sleep(self.fan_level_delay)

    def set_multiple_fan_levels(self, zone_list: List[int], level: int) -> None:
        """Set the fan level in multiple IPMI zones. Could raise several exceptions in case of invalid parameters.
        Args:
            zone_list (List[int]): List of IPMI zones
            level (int): fan level in % (0-100)
        Raises:
            ValueError: invalid input parameter
            FileNotFoundError: ipmitool command cannot be found
            RuntimeError: ipmitool execution problem (e.g. non-root user, incompatible IPMI system/motherboard)
        """
        # Validate zone parameters
        for zone in zone_list:
            if zone not in range(0, 101):
                raise ValueError(f'Invalid value: zone ({zone}).')
        # Validate level parameter (must be in the interval [0..100%])
        if level not in range(0, 101):
            raise ValueError(f'Invalid value: level ({level}).')
        # Set the new IPMI fan level in the specific zone
        try:
            for zone in zone_list:
                self._exec_ipmitool(['raw', '0x30', '0x70', '0x66', '0x01', f'0x{zone:02x}', f'0x{level:02x}'])
        except (FileNotFoundError, RuntimeError) as e:
            raise e
        # Give time for IPMI and fans to spin up/down.
        time.sleep(self.fan_level_delay)

    def get_fan_level(self, zone: int) -> int:
        """Get the current fan level in a specific IPMI zone. Raise an exception in case of invalid parameters.
        Args:
            zone (int): fan zone (CPU_ZONE, HD_ZONE)
        Returns:
            level (int): fan level in % (0-100)
        Raises:
            ValueError: invalid input parameter
            FileNotFoundError: ipmitool command cannot be found
            RuntimeError: ipmitool execution problem (e.g. non-root user, incompatible IPMI system/motherboard)
        """
        r: subprocess.CompletedProcess  # result of the executed process
        level: int                      # Level

        # Validate zone parameter
        if zone not in range(0, 101):
            raise ValueError(f'Invalid value: zone ({zone}).')
        # Get the new IPMI fan level in the specific zone
        try:
            r = self._exec_ipmitool(['raw', '0x30', '0x70', '0x66', '0x00', f'0x{zone:x}'])
            level = int(r.stdout, 16)
        except (FileNotFoundError, RuntimeError) as e:
            raise e
        return level


# End.



#
#   gpuzone.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.GpuZone() class implementation.
#
import subprocess
import time
import re
from configparser import ConfigParser
from typing import List
from smfc.fancontroller import FanController
from smfc.ipmi import Ipmi
from smfc.log import Log


class GpuZone(FanController):
    """Class for GPU zone fan control."""

    # GpuZone specific parameters.
    gpu_device_ids: List[int]           # GPU device IDs (indexes)
    nvidia_smi_path: str                # Path for `nvidia-smi` command
    nvidia_smi_called: float            # Timestamp when `nvidia-smi` command executed
    gpu_temperature: List[float]        # List of GPU temperatures

    # Constant values for the configuration parameters.
    CS_GPU_ZONE: str = 'GPU zone'
    CV_GPU_ZONE_ENABLED: str = 'enabled'
    CV_GPU_IPMI_ZONE: str = 'ipmi_zone'
    CV_GPU_ZONE_TEMP_CALC: str = 'temp_calc'
    CV_GPU_ZONE_STEPS: str = 'steps'
    CV_GPU_ZONE_SENSITIVITY: str = 'sensitivity'
    CV_GPU_ZONE_POLLING: str = 'polling'
    CV_GPU_ZONE_MIN_TEMP: str = 'min_temp'
    CV_GPU_ZONE_MAX_TEMP: str = 'max_temp'
    CV_GPU_ZONE_MIN_LEVEL: str = 'min_level'
    CV_GPU_ZONE_MAX_LEVEL: str = 'max_level'
    CV_GPU_ZONE_GPU_IDS: str = 'gpu_device_ids'
    CV_GPU_ZONE_NVIDIA_SMI_PATH: str = 'nvidia_smi_path'

    def __init__(self, log: Log, ipmi: Ipmi, config: ConfigParser) -> None:
        """Initialize the GpuZone class. Abort in case of configuration errors.
        Args:
            log (Log): reference to a Log class instance
            ipmi (Ipmi): reference to an Ipmi class instance
            config (configparser.ConfigParser): reference to the configuration (default=None)
        Raises:
            ValueError: invalid parameters
        """
        gpu_id_list: str    # String for gpu_device_ids=
        count: int          # GPU count.

        # Save and validate GpuZone class-specific parameters.
        gpu_id_list = config[self.CS_GPU_ZONE].get(self.CV_GPU_ZONE_GPU_IDS, '0')
        gpu_id_list = re.sub(' +', ' ', gpu_id_list.strip())
        try:
            self.gpu_device_ids = [int(s) for s in gpu_id_list.split(',' if ',' in gpu_id_list else ' ')]
        except ValueError as e:
            raise e
        for gid in self.gpu_device_ids:
            if gid not in range(0, 101):
                raise ValueError(f'invalid value: {self.CV_GPU_ZONE_GPU_IDS}={gpu_id_list}.')
        count = len(self.gpu_device_ids)
        self.nvidia_smi_path = config[GpuZone.CS_GPU_ZONE].get(GpuZone.CV_GPU_ZONE_NVIDIA_SMI_PATH,
                                                              '/usr/bin/nvidia-smi')
        self.nvidia_smi_called = 0

        # Initialize FanController class.
        super().__init__(log, ipmi,
            config[GpuZone.CS_GPU_ZONE].get(GpuZone.CV_GPU_IPMI_ZONE, fallback=f'{Ipmi.HD_ZONE}'),
            GpuZone.CS_GPU_ZONE, count,
            config[GpuZone.CS_GPU_ZONE].getint(GpuZone.CV_GPU_ZONE_TEMP_CALC, fallback=FanController.CALC_AVG),
            config[GpuZone.CS_GPU_ZONE].getint(GpuZone.CV_GPU_ZONE_STEPS, fallback=5),
            config[GpuZone.CS_GPU_ZONE].getfloat(GpuZone.CV_GPU_ZONE_SENSITIVITY, fallback=2),
            config[GpuZone.CS_GPU_ZONE].getfloat(GpuZone.CV_GPU_ZONE_POLLING, fallback=2),
            config[GpuZone.CS_GPU_ZONE].getfloat(GpuZone.CV_GPU_ZONE_MIN_TEMP, fallback=40),
            config[GpuZone.CS_GPU_ZONE].getfloat(GpuZone.CV_GPU_ZONE_MAX_TEMP, fallback=70),
            config[GpuZone.CS_GPU_ZONE].getint(GpuZone.CV_GPU_ZONE_MIN_LEVEL, fallback=35),
            config[GpuZone.CS_GPU_ZONE].getint(GpuZone.CV_GPU_ZONE_MAX_LEVEL, fallback=100)
        )

        # Print configuration in CONFIG log level (or higher).
        if self.log.log_level >= Log.LOG_CONFIG:
            self.log.msg(Log.LOG_CONFIG, f'   {self.CV_GPU_ZONE_GPU_IDS} = {self.gpu_device_ids}')
            self.log.msg(Log.LOG_CONFIG, f'   {self.CV_GPU_ZONE_NVIDIA_SMI_PATH} = {self.nvidia_smi_path}')

    def _exec_nvidia_smi(self, arguments: List[str]) -> subprocess.CompletedProcess:
        """Execution of the `nvidia-smi` command.
            Args:
                arguments (List[str]): list of argument of `nvidia-smi` command
            Raises:
                FileNotFoundError: command not found
        """
        r: subprocess.CompletedProcess  # Result of the executed process
        args: List[str] = []            # List of arguments

        # Execute `nvidia-smi` command.
        try:
            args.append(self.nvidia_smi_path)
            args.extend(arguments)
            r = subprocess.run(args, check=False, capture_output=True, text=True)
        except FileNotFoundError as e:
            raise e
        return r

    def _get_nth_temp(self, index: int) -> float:
        """Get the temperature of the nth element in the GPU device list.
        Args:
            index (int): index in GPU device list
        Returns:
            float: temperature value
        Raises:
            FileNotFoundError:  file or command cannot be found
            ValueError:         invalid temperature value
            IndexError:         invalid index
        """
        current_time = time.monotonic()
        if (current_time - self.nvidia_smi_called) >= self.polling:
            r: subprocess.CompletedProcess  # result of the executed process

            r = self._exec_nvidia_smi(['--query-gpu=temperature.gpu', '--format=csv,noheader,nounits'])
            self.nvidia_smi_called = current_time
            temp_list = r.stdout.splitlines()
            self.gpu_temperature = []
            for gid in self.gpu_device_ids:
                self.gpu_temperature.append(int(temp_list[gid]))

        return self.gpu_temperature[index]


# End.



#!/usr/bin/env python3
#
#   test_08_cmd.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.cmd() class.
#
from mock import MagicMock
from pytest_mock import MockerFixture
from smfc import main


# pylint: disable=too-few-public-methods
class TestMain:
    """Unit test for smfc.main() function."""

    def test_main(self, mocker:MockerFixture) -> None:
        """This is a unit test for function Service.main()."""
        mock_service_run = MagicMock()
        mocker.patch('smfc.Service.run', mock_service_run)
        main()
        mock_service_run.assert_called_once()
# pylint: enable=too-few-public-methods


# End.



#
#   log.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.Log() class implementation.
#
import sys
import syslog
from typing import Callable


class Log:
    """Log class. This class can send log messages considering different log levels and different log outputs."""

    # Configuration parameters.
    log_level: int                      # Log level
    log_output: int                     # Log output
    msg: Callable[[int, str], None]     # Function reference to the log function (based on log output)

    # Constants for log levels.
    LOG_NONE: int = 0
    LOG_ERROR: int = 1
    LOG_CONFIG: int = 2
    LOG_INFO: int = 3
    LOG_DEBUG: int = 4

    # Constants for log outputs.
    LOG_STDOUT: int = 0
    LOG_STDERR: int = 1
    LOG_SYSLOG: int = 2

    def __init__(self, log_level: int, log_output: int) -> None:
        """Initialize Log class with log output and log level.
        Args:
            log_level (int): user defined log level (LOG_NONE, LOG_ERROR, LOG_CONFIG, LOG_INFO, LOG_DEBUG)
            log_output (int): user defined log output (LOG_STDOUT, LOG_STDERR, LOG_SYSLOG)
        Raises:
            ValueError: invalid input parameters
        """
        # Setup log configuration.
        if log_level not in {Log.LOG_NONE, Log.LOG_ERROR, Log.LOG_CONFIG, Log.LOG_INFO, Log.LOG_DEBUG}:
            raise ValueError(f'Invalid log level value ({log_level})')
        self.log_level = log_level
        if log_output not in {Log.LOG_STDOUT, Log.LOG_STDERR, Log.LOG_SYSLOG}:
            raise ValueError(f'Invalid log output value ({log_output})')
        self.log_output = log_output
        if self.log_output == Log.LOG_STDOUT:
            self.msg = self.msg_to_stdout
        elif self.log_output == Log.LOG_STDERR:
            self.msg = self.msg_to_stderr
        else:
            self.msg = self.msg_to_syslog
            syslog.openlog('smfc.service', facility=syslog.LOG_DAEMON)

    @staticmethod
    def map_to_syslog(level: int) -> int:
        """Map log level to syslog values.
            Args:
                level (int): log level (LOG_ERROR, LOG_CONFIG, LOG_INFO, LOG_DEBUG)
            Returns:
                int: syslog log level
            """
        syslog_level = syslog.LOG_ERR
        if level in (Log.LOG_CONFIG, Log.LOG_INFO):
            syslog_level = syslog.LOG_INFO
        elif level == Log.LOG_DEBUG:
            syslog_level = syslog.LOG_DEBUG
        return syslog_level

    @staticmethod
    def level_to_str(level: int) -> str:
        """Convert a log level to a string.
            Args:
                level (int): log level (LOG_ERROR, LOG_CONFIG, LOG_INFO, LOG_DEBUG)
            Returns:
                str: log level string
            """
        string = 'NONE'
        if level == Log.LOG_ERROR:
            string = 'ERROR'
        if level == Log.LOG_CONFIG:
            string = 'CONFIG'
        if level == Log.LOG_INFO:
            string = 'INFO'
        elif level == Log.LOG_DEBUG:
            string = 'DEBUG'
        return string

    def msg_to_syslog(self, level: int, msg: str) -> None:
        """Print a log message to syslog.
        Args:
            level (int): log level (LOG_ERROR, LOG_CONFIG, LOG_INFO, LOG_DEBUG)
            msg (str): log message
        """
        if level is not Log.LOG_NONE:
            if level <= self.log_level:
                syslog.syslog(self.map_to_syslog(level), msg)

    def msg_to_stdout(self, level: int, msg: str) -> None:
        """Print a log message to stdout.
        Args:
            level (int): log level (LOG_ERROR, LOG_CONFIG, LOG_INFO, LOG_DEBUG)
            msg (str):  log message
        """
        if level is not Log.LOG_NONE:
            if level <= self.log_level:
                print(f'{self.level_to_str(level)}: {msg}', flush=True, file=sys.stdout)

    def msg_to_stderr(self, level: int, msg: str) -> None:
        """Print a log message to stderr.

        Args:
            level (int): log level (LOG_ERROR, LOG_CONFIG, LOG_INFO, LOG_DEBUG)
            msg (str):  log message
        """
        if level is not Log.LOG_NONE:
            if level <= self.log_level:
                print(f'{self.level_to_str(level)}: {msg}', flush=True, file=sys.stderr)

# End.



#!/usr/bin/env python3
#
#   test_07_service.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.Service() class.
#
from argparse import Namespace
import sys
import time
from configparser import ConfigParser
import pytest
from pyudev import Context
from mock import MagicMock
from pytest_mock import MockerFixture
from smfc import Log, Ipmi, CpuZone, HdZone, GpuZone, Service, FanController
from .test_00_data import TestData, MockedContextError, MockedContextGood

class TestService:
    """Unit test for smfc.Service() class"""

    sleep_counter: int

    @pytest.mark.parametrize("ipmi, log, error", [
        (True, True,   'Service.exit_func() 1'),
        (False, False, 'Service.exit_func() 2')
    ])
    def test_exit_func(self, mocker:MockerFixture, ipmi: bool, log: bool, error: str) -> None:
        """Positive unit test for Service.exit_func() method. It contains the following steps:
            - mock atexit.unregister(), Ipmi.set_fan_level(), Log.msg_to_stdout() functions
            - execute Service.exit_func()
            - ASSERT: if mocked functions not called expected times
        """
        mock_atexit_unregister = MagicMock()
        mocker.patch('atexit.unregister', mock_atexit_unregister)
        mock_ipmi_set_fan_level = MagicMock()
        mocker.patch('smfc.Ipmi.set_fan_level', mock_ipmi_set_fan_level)
        mock_log_msg = MagicMock()
        mocker.patch('smfc.Log.msg_to_stdout', mock_log_msg)
        service = Service()
        if log:
            service.log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        if ipmi:
            service.ipmi = Ipmi.__new__(Ipmi)
        service.exit_func()
        assert mock_atexit_unregister.call_count == 1, error
        if ipmi:
            assert mock_ipmi_set_fan_level.call_count == 2, error
            if log:
                assert mock_log_msg.call_count == 1, error

    @pytest.mark.parametrize("module_list, cpuzone, hdzone, gpuzone, standby, error", [
        ("something\ncoretemp\n",           True,  False, False, False, "Service.check_dependencies() 1"),
        ("something\nk10temp\n",            True,  False, True,  False, "Service.check_dependencies() 2"),
        ("coretemp\nsomething\nk10temp\n",  True,  False, False, False, "Service.check_dependencies() 3"),
        ("something\ndrivetemp\n",          False, True,  True,  False, "Service.check_dependencies() 4"),
        ("something\ndrivetemp\n",          False, True,  False, True,  "Service.check_dependencies() 5"),
        ("something\n",                     False, True,  False, False, "Service.check_dependencies() 6"),
        ("something\ndrivetemp\nx",         False, True,  True,  True,  "Service.check_dependencies() 7"),
        ("coretemp\ndrivetemp\n",           True,  True,  False, True,  "Service.check_dependencies() 8")
    ])
    def test_check_dependencies_p(self, mocker: MockerFixture, module_list: str, cpuzone: bool, hdzone: bool,
                                  gpuzone: bool, standby: bool, error: str):
        """Positive unit test for Service.check_dependencies() method. It contains the following steps:
            - mock print(), argparse.ArgumentParser._print_message() and builtins.open() functions
            - execute Service.check_dependencies()
            - ASSERT: if returns an error message
        """

        def mocked_open(path: str, *args, **kwargs):
            return original_open(modules, *args, **kwargs) if path == "/proc/modules" else \
                   original_open(path, *args, **kwargs)

        my_td = TestData()
        ipmi_command = my_td.create_ipmi_command()
        modules = my_td.create_text_file(module_list)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        original_open = open
        mock_open = MagicMock(side_effect=mocked_open)
        mocker.patch('builtins.open', mock_open)

        service = Service()
        service.config = ConfigParser()
        service.config[Ipmi.CS_IPMI] = {}
        service.config[Ipmi.CS_IPMI][Ipmi.CV_IPMI_COMMAND] = ipmi_command

        service.cpu_zone_enabled = cpuzone
        service.config[CpuZone.CS_CPU_ZONE] = {}
        service.config[CpuZone.CS_CPU_ZONE][CpuZone.CV_CPU_ZONE_ENABLED] = '1' if cpuzone else '0'

        service.hd_zone_enabled = hdzone
        service.config[HdZone.CS_HD_ZONE] = {}
        service.config[HdZone.CS_HD_ZONE][HdZone.CV_HD_ZONE_ENABLED] = '1' if hdzone else '0'

        service.gpu_zone_enabled = gpuzone
        service.config[GpuZone.CS_GPU_ZONE] = {}
        service.config[GpuZone.CS_GPU_ZONE][GpuZone.CV_GPU_ZONE_ENABLED] = '1' if gpuzone else '0'

        smartctl_cmd = my_td.create_command_file('echo "ACTIVE"')
        if hdzone:
            service.config[HdZone.CS_HD_ZONE][HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED] = '1' if standby else '0'
            service.config[HdZone.CS_HD_ZONE][HdZone.CV_HD_ZONE_SMARTCTL_PATH] = smartctl_cmd
        assert service.check_dependencies() == '', error
        mocker.patch('builtins.open', original_open)
        del my_td

    @pytest.mark.parametrize("error", ["Service.check_dependencies() 9"])
    def test_check_dependecies_n(self, mocker:MockerFixture, error: str):
        """Negative unit test fot Service.check_dependencies() method. It contains the following steps:
            - mock print() and builtins.open() functions
            - execute Service.check_dependencies()
            - ASSERT: if it didn't return the specific error message
        """

        def mocked_open(path: str, *args, **kwargs):
            return original_open(modules, *args, **kwargs) if path == "/proc/modules" else \
                   original_open(path, *args, **kwargs)

        my_td = TestData()
        ipmi_command = my_td.create_ipmi_command()
        modules = my_td.create_text_file("coretemp\ndrivetemp\n")
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_open = MagicMock(side_effect=mocked_open)
        original_open = open
        mocker.patch('builtins.open', mock_open)
        service = Service()
        service.config = ConfigParser()
        service.config[Ipmi.CS_IPMI] = {}
        service.config[Ipmi.CS_IPMI][Ipmi.CV_IPMI_COMMAND] = ipmi_command
        service.cpu_zone_enabled = True
        service.config[CpuZone.CS_CPU_ZONE] = {}
        service.config[CpuZone.CS_CPU_ZONE][CpuZone.CV_CPU_ZONE_ENABLED] = '1'
        service.hd_zone_enabled = True
        service.config[HdZone.CS_HD_ZONE] = {}
        service.config[HdZone.CS_HD_ZONE][HdZone.CV_HD_ZONE_ENABLED] = '1'
        service.gpu_zone_enabled = True
        service.config[GpuZone.CS_GPU_ZONE] = {}
        service.config[GpuZone.CS_GPU_ZONE][GpuZone.CV_GPU_ZONE_ENABLED] = '1'
        smartctl_cmd = my_td.create_command_file('echo "ACTIVE"')
        service.config[HdZone.CS_HD_ZONE][HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED] = '1'
        service.config[HdZone.CS_HD_ZONE][HdZone.CV_HD_ZONE_SMARTCTL_PATH] = smartctl_cmd

        # Check if `smartctl` command is not available.
        my_td.delete_file(smartctl_cmd)
        error_str = service.check_dependencies()
        assert error_str.find('smartctl') != -1, error

        # Check if `drivetemp` is not on the module list.
        modules = my_td.create_text_file('coretemp something')
        error_str = service.check_dependencies()
        assert error_str.find('drivetemp') != -1, error

        # Check if `coretemp` is not on module list.
        modules = my_td.create_text_file('drivetemp something')
        error_str = service.check_dependencies()
        assert error_str.find('coretemp') != -1, error

        # Check if `ipmitool` is not available.
        my_td.delete_file(ipmi_command)
        error_str = service.check_dependencies()
        assert error_str.find('ipmitool') != -1, error
        del my_td

    @pytest.mark.parametrize("command_line, exit_code, error", [
        ('-h',                                                      0, 'Service.run() 1'),
        ('-v',                                                      0, 'Service.run() 2'),
        ('-l 10',                                                   2, 'Service.run() 3'),
        ('-o 9',                                                    2, 'Service.run() 4'),
        ('-o 1 -l 10',                                              2, 'Service.run() 5'),
        ('-o 9 -l 1',                                               2, 'Service.run() 6'),
        ('-o 0 -l 3 -c &.txt',                                      6, 'Service.run() 7'),
        ('-o 0 -l 3 -c ./nonexistent_folder/nonexistent_file.conf', 6, 'Service.run() 8')
    ])
    def test_run_026n(self, mocker:MockerFixture, command_line: str, exit_code: int, error: str):
        """Negative unit test for Service.run() method. It contains the following steps:
            - mock print(), argparse.ArgumentParser._print_message() functions
            - execute Service.run()
            - ASSERT: if sys.exit() did not return code 0 (-h -v), 2 (invalid arguments), 6 (invalid configuration file)
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker_argumentparser_print = MagicMock()
        mocker.patch('argparse.ArgumentParser._print_message', mocker_argumentparser_print)
        sys.argv = ('smfc ' + command_line).split()
        service = Service()
        with pytest.raises(SystemExit) as cm:
            service.run()
        assert cm.value.code == exit_code, error

    @pytest.mark.parametrize("level, output, exit_code, error", [
        (10, 0, 5, 'Service.run() 9'),
        (0,  9, 5, 'Service.run() 10')
    ])
    def test_run_5n(self, mocker:MockerFixture, level: int, output: int, exit_code: int, error: str):
        """Negative unit test for Service.run() method. It contains the following steps:
            - mock print(), argparse.ArgumentParser.parse_args() functions
            - execute Service.run()
            - ASSERT: if sys.exit() did not return code 5 (log initialization error)
        """
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_parser_parse_args = MagicMock()
        mocker.patch('argparse.ArgumentParser.parse_args', mock_parser_parse_args)
        mock_parser_parse_args.return_value = Namespace(config_file="smfc.conf", ne=False, s=False, l=level, o=output)
        service = Service()
        with pytest.raises(SystemExit) as cm:
            service.run()
        assert cm.value.code == exit_code, error

    @pytest.mark.parametrize("exit_code, error", [
        (7, 'Service.run() 11')
    ])
    def test_run_7n(self, mocker:MockerFixture, exit_code: int, error: str):
        """Negative unit test for Service.run() method. It contains the following steps:
            - mock print(), argparse.ArgumentParser.parse_args(), smfc.Service.check_dependencies() functions
            - execute Service.run()
            - ASSERT: if sys.exit() did not return code 7 (check dependency error)
        """
        my_td = TestData()
        my_config = ConfigParser()
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '0'
        }
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: '0'
        }
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: '0'
        }
        conf_file = my_td.create_config_file(my_config)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_parser_parse_args = MagicMock()
        mocker.patch('argparse.ArgumentParser.parse_args', mock_parser_parse_args)
        mock_parser_parse_args.return_value = Namespace(config_file=conf_file, ne=True, nd=False, s=False, l=0, o=0)
        mock_check_dependencies = MagicMock()
        mock_check_dependencies.return_value = "ERROR"
        mocker.patch('smfc.Service.check_dependencies', mock_check_dependencies)
        service = Service()
        with pytest.raises(SystemExit) as cm:
            service.run()
        assert cm.value.code == exit_code, error
        del my_td

    @pytest.mark.parametrize("ipmi_command, mode_delay, level_delay, exit_code, error", [
        ('NON_EXIST', 0,  0,  8,  'Service.run() 12'),
        ('GOOD',      -1, 0,  8,  'Service.run() 13'),
        ('GOOD',      0,  -1, 8,  'Service.run() 14'),
        ('BAD',       0,  0,  8,  'Service.run() 15'),
        ('GOOD',      0,  0,  10, 'Service.run() 16')
    ])
    def test_run_810n(self, mocker:MockerFixture, ipmi_command: str, mode_delay: int, level_delay: int, exit_code: int,
                      error: str):
        """Negative unit test for Service.run() method. It contains the following steps:
            - mock print(), pyudev.Context.__init__() functions
            - execute Service.run()
            - ASSERT: if sys.exit() did not return code 8 (Ipmi initialization error) or 10 (no enabled zone)
        """
        my_td = TestData()
        my_config = ConfigParser()
        if ipmi_command == 'NON_EXIST':
            ipmi_command = './non-existent-dir/non-existent-file'
        if ipmi_command == 'BAD':
            ipmi_command = my_td.create_command_file()
        if ipmi_command == 'GOOD':
            ipmi_command = my_td.create_ipmi_command()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: ipmi_command,
            Ipmi.CV_IPMI_FAN_MODE_DELAY: str(mode_delay),
            Ipmi.CV_IPMI_FAN_LEVEL_DELAY: str(level_delay),
        }
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '0'
        }
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: '0'
        }
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: '0'
        }
        conf_file = my_td.create_config_file(my_config)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker.patch('pyudev.Context.__init__', MockedContextGood.__init__)
        sys.argv = ('smfc.py -o 0 -nd -ne -c ' + conf_file).split()
        service = Service()
        with pytest.raises(SystemExit) as cm:
            service.run()
        assert cm.value.code == exit_code, error
        del my_td

    @pytest.mark.parametrize("exit_code, error", [
        (9, 'Service.run() 17')
    ])
    def test_run_9n(self, mocker:MockerFixture, exit_code: int, error: str):
        """Negative unit test for Service.run() method. It contains the following steps:
            - mock print(), pyudev.Context.__init__() functions
            - execute Service.run()
            - ASSERT: if sys.exit() did not return code 9 (pyudev.Context() init error)
        """
        my_td = TestData()
        my_config = ConfigParser()
        ipmi_command = my_td.create_ipmi_command()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: ipmi_command,
            Ipmi.CV_IPMI_FAN_MODE_DELAY: '0',
            Ipmi.CV_IPMI_FAN_LEVEL_DELAY: '0',
        }
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: '0'
        }
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: '0'
        }
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: '0'
        }
        conf_file = my_td.create_config_file(my_config)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mocker.patch('pyudev.Context.__init__', MockedContextError.__init__)
        sys.argv = ('smfc.py -o 0 -ne -nd -c ' + conf_file).split()
        service = Service()
        with pytest.raises(SystemExit) as cm:
            service.run()
        assert cm.value.code == exit_code, error
        del my_td

    @pytest.mark.parametrize("cpuzone, hdzone, gpuzone, exit_code, error", [
        (1, 0, 1, 100, 'Service.run() 18'),
        (0, 1, 0, 100, 'Service.run() 19'),
        (1, 0, 1, 100, 'Service.run() 20')
    ])
    def test_run_100p(self, mocker:MockerFixture, cpuzone: int, hdzone: int, gpuzone: int, exit_code: int, error: str):
        """Positive unit test for Service.run() method. It contains the following steps:
            - mock print(), time.sleep() functions
            - execute smfc.run()
            - The main loop will be executed 10 times then exit with code 100
        """

        # pylint: disable=unused-argument
        def mocked_sleep(*args):
            """Mocked time.sleep() function. Exists at the 10th call."""
            self.sleep_counter += 1
            if self.sleep_counter >= 10:
                sys.exit(100)

        def mocked_cpuzone_init(self, log: Log, udevc: Context, ipmi: Ipmi, config: ConfigParser) -> None:
            nonlocal my_td
            self.hwmon_path = my_td.cpu_files
            count = len(my_td.cpu_files)
            FanController.__init__(self, log, ipmi, f'{Ipmi.CPU_ZONE}', CpuZone.CS_CPU_ZONE, count, 1, 5,
                                  5, 0, 30, 60, 35, 100)

        def mocked_hdzone_init(self, log: Log, udevc: Context, ipmi: Ipmi, config: ConfigParser, sudo: bool) -> None:
            nonlocal my_td
            nonlocal cmd_smart
            self.hd_device_names = my_td.hd_name_list
            self.hwmon_path = my_td.hd_files
            count = len(my_td.hd_files)
            self.sudo=sudo
            FanController.__init__(self, log, ipmi, f'{Ipmi.HD_ZONE}', HdZone.CS_HD_ZONE, count, 1, 5,
                                  2, 0, 32, 46, 35, 100)
            self.smartctl_path = cmd_smart
            self.standby_guard_enabled = True
            self.standby_hd_limit = 1
            self.standby_array_states = [False] * self.count
            self.standby_flag = False
            self.standby_change_timestamp = time.monotonic()

        def mocked_gpuzone_init(self, log: Log, ipmi: Ipmi, config: ConfigParser) -> None:
            nonlocal my_td
            nonlocal cmd_nvidia
            self.gpu_device_ids = [0]
            count = 1
            self.nvidia_smi_path = cmd_nvidia
            self.nvidia_smi_called = 0
            FanController.__init__(self, log, ipmi, '2', GpuZone.CS_GPU_ZONE, count, 1, 5,
                                  2, 0, 45, 70, 35, 100)
        # pragma pylint: enable=unused-argument

        my_td = TestData()
        # Force mode initial fan mode 0 for setting new FULL mode during the test.
        cmd_ipmi = my_td.create_command_file('echo "0"')
        cmd_smart = my_td.create_smart_command()
        #                     create_command_file('echo "ACTIVE"'))
        cmd_nvidia = my_td.create_nvidia_smi_command(1)
        my_td.create_cpu_data(1)
        my_td.create_hd_data(8)
        my_config = ConfigParser()
        my_config[Ipmi.CS_IPMI] = {
            Ipmi.CV_IPMI_COMMAND: cmd_ipmi,
            Ipmi.CV_IPMI_FAN_MODE_DELAY: '0',
            Ipmi.CV_IPMI_FAN_LEVEL_DELAY: '0'
        }
        my_config[CpuZone.CS_CPU_ZONE] = {
            CpuZone.CV_CPU_ZONE_ENABLED: str(cpuzone),
            CpuZone.CV_CPU_ZONE_TEMP_CALC: '1',
            CpuZone.CV_CPU_ZONE_STEPS: '5',
            CpuZone.CV_CPU_ZONE_SENSITIVITY: '5',
            CpuZone.CV_CPU_ZONE_POLLING: '0',
            CpuZone.CV_CPU_ZONE_MIN_TEMP: '30',
            CpuZone.CV_CPU_ZONE_MAX_TEMP: '60',
            CpuZone.CV_CPU_ZONE_MIN_LEVEL: '35',
            CpuZone.CV_CPU_ZONE_MAX_LEVEL: '100',
        }
        my_config[HdZone.CS_HD_ZONE] = {
            HdZone.CV_HD_ZONE_ENABLED: str(hdzone),
            HdZone.CV_HD_ZONE_TEMP_CALC: '1',
            HdZone.CV_HD_ZONE_STEPS: '4',
            HdZone.CV_HD_ZONE_SENSITIVITY: '2',
            HdZone.CV_HD_ZONE_POLLING: '0',
            HdZone.CV_HD_ZONE_MIN_TEMP: '30',
            HdZone.CV_HD_ZONE_MAX_TEMP: '45',
            HdZone.CV_HD_ZONE_MIN_LEVEL: '35',
            HdZone.CV_HD_ZONE_MAX_LEVEL: '100',
            HdZone.CV_HD_ZONE_HD_NAMES: my_td.hd_names,
            HdZone.CV_HD_ZONE_SMARTCTL_PATH: cmd_smart,
            HdZone.CV_HD_ZONE_STANDBY_GUARD_ENABLED: '1',
            HdZone.CV_HD_ZONE_STANDBY_HD_LIMIT: '2'
        }
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: str(gpuzone),
            GpuZone.CV_GPU_IPMI_ZONE: '2',
            GpuZone.CV_GPU_ZONE_TEMP_CALC: '1',
            GpuZone.CV_GPU_ZONE_STEPS: '4',
            GpuZone.CV_GPU_ZONE_SENSITIVITY: '2',
            GpuZone.CV_GPU_ZONE_POLLING: '0',
            GpuZone.CV_GPU_ZONE_MIN_TEMP: '45',
            GpuZone.CV_GPU_ZONE_MAX_TEMP: '70',
            GpuZone.CV_GPU_ZONE_MIN_LEVEL: '35',
            GpuZone.CV_GPU_ZONE_MAX_LEVEL: '100',
            GpuZone.CV_GPU_ZONE_GPU_IDS: '0',
            GpuZone.CV_GPU_ZONE_NVIDIA_SMI_PATH: cmd_nvidia,
        }
        conf_file = my_td.create_config_file(my_config)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_time_sleep = MagicMock()
        mock_time_sleep.side_effect = mocked_sleep
        mocker.patch('time.sleep', mock_time_sleep)
        mocker.patch('pyudev.Context.__init__', MockedContextGood.__init__)
        mocker.patch('smfc.CpuZone.__init__', mocked_cpuzone_init)
        mocker.patch('smfc.HdZone.__init__', mocked_hdzone_init)
        mocker.patch('smfc.GpuZone.__init__', mocked_gpuzone_init)
        self.sleep_counter = 0
        sys.argv = ('smfc.py -o 0 -l 4 -ne -nd -c ' + conf_file).split()
        service = Service()
        with pytest.raises(SystemExit) as cm:
            service.run()
        assert cm.value.code == exit_code, error
        del my_td


# End.



#
#   cmd.py (C) 2020-2025, Peter Sulyok
#   smfc package: Super Micro fan control for Linux (home) servers.
#   smfc.main() function implementation, command-line interface.
#
from smfc.service import Service


def main() -> None:
    """Entry point of the `smfc` program."""
    service = Service()
    service.run()


if __name__ == '__main__':
    main()

# End.



#!/usr/bin/env python3
#
#   test_06_gpuzone.py (C) 2021-2025, Peter Sulyok
#   Unit tests for smfc.GpuZone() class.
#
import subprocess
from configparser import ConfigParser
from typing import List, Any
import pytest
from mock import MagicMock
from pytest_mock import MockerFixture
from smfc import Log, Ipmi, FanController, GpuZone
from .test_00_data import TestData

class TestGpuZone:
    """Unit test class for smfc.GpuZone() class"""

    @pytest.mark.parametrize(
        "count, ipmi_zone, gpu_device_ids, temp_calc, steps, sensitivity, polling, min_temp, max_temp, min_level, "
        "max_level, error", [
        (1, '0', '0',           FanController.CALC_MIN, 4, 2, 2, 32, 48, 35, 100, 'GpuZone.__init__() 1'),
        (2, '1', '0, 1',        FanController.CALC_AVG, 4, 2, 2, 32, 48, 35, 100, 'GpuZone.__init__() 2'),
        (4, '2', '0, 1, 2, 3',  FanController.CALC_AVG, 4, 2, 2, 32, 48, 35, 100, 'GpuZone.__init__() 3')
    ])
    def test_init_p1(self, mocker: MockerFixture, count: int, ipmi_zone: str, gpu_device_ids: str, temp_calc: int,
                     steps: int, sensitivity: float, polling: float, min_temp: float, max_temp: float, min_level: int,
                     max_level: int, error: str):
        """Positive unit test for GpuZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Devices.from_device_file(), pyudev.Device, smfc.FanController.get_hwmon_path()
            - initialize a Config, Log, Ipmi, and GpuZone classes
            - ASSERT: if the GpuZone class attributes are different from values passed to __init__
        """
        my_td = TestData()
        nvidia_smi_cmd = my_td.create_nvidia_smi_command(count)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        my_config = ConfigParser()
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: '1',
            GpuZone.CV_GPU_IPMI_ZONE: ipmi_zone,
            GpuZone.CV_GPU_ZONE_TEMP_CALC: str(temp_calc),
            GpuZone.CV_GPU_ZONE_STEPS: str(steps),
            GpuZone.CV_GPU_ZONE_SENSITIVITY: str(sensitivity),
            GpuZone.CV_GPU_ZONE_POLLING: str(polling),
            GpuZone.CV_GPU_ZONE_MIN_TEMP: str(min_temp),
            GpuZone.CV_GPU_ZONE_MAX_TEMP: str(max_temp),
            GpuZone.CV_GPU_ZONE_MIN_LEVEL: str(min_level),
            GpuZone.CV_GPU_ZONE_MAX_LEVEL: str(max_level),
            GpuZone.CV_GPU_ZONE_GPU_IDS: gpu_device_ids,
            GpuZone.CV_GPU_ZONE_NVIDIA_SMI_PATH: nvidia_smi_cmd
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_gpuzone = GpuZone(my_log, my_ipmi, my_config)
        assert my_gpuzone.ipmi_zone == [int(s) for s in ipmi_zone.split(',' if ',' in ipmi_zone else ' ')], error
        assert my_gpuzone.name == GpuZone.CS_GPU_ZONE, error
        assert my_gpuzone.count == count, error
        assert my_gpuzone.temp_calc == temp_calc, error
        assert my_gpuzone.steps == steps, error
        assert my_gpuzone.sensitivity == sensitivity, error
        assert my_gpuzone.polling == polling, error
        assert my_gpuzone.min_temp == min_temp, error
        assert my_gpuzone.max_temp == max_temp, error
        assert my_gpuzone.min_level == min_level, error
        assert my_gpuzone.max_level == max_level, error
        assert my_gpuzone.gpu_device_ids == [int(s) for s in gpu_device_ids.
                                             split(',' if ',' in gpu_device_ids else ' ')], error
        assert my_gpuzone.nvidia_smi_path == nvidia_smi_cmd, error
        del my_td

    @pytest.mark.parametrize("error", [
        'GpuZone.__init__() 4'
    ])
    def test_init_p2(self, mocker: MockerFixture, error: str):
        """Positive unit test for GpuZone.__init__() method. It contains the following steps:
            - mock print(), smfc.GpuZone._exec_nvidia_smi()
            - initialize a Config, Log, Ipmi, and GpuZone classes
            - ASSERT: if the GpuZone class attributes are different from the default configuration values
        """
        count = 1
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        mock_exec_nvidia_smi = MagicMock()
        mock_exec_nvidia_smi.return_value = subprocess.CompletedProcess([], 0, stdout='40\n')
        mocker.patch('smfc.GpuZone._exec_nvidia_smi', mock_exec_nvidia_smi)
        my_config = ConfigParser()
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: '1',
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        my_gpuzone = GpuZone(my_log, my_ipmi, my_config)
        assert my_gpuzone.log == my_log, error
        assert my_gpuzone.ipmi == my_ipmi
        assert my_gpuzone.ipmi_zone == [Ipmi.HD_ZONE], error
        assert my_gpuzone.name == GpuZone.CS_GPU_ZONE, error
        assert my_gpuzone.count == count, error
        assert my_gpuzone.temp_calc == FanController.CALC_AVG, error
        assert my_gpuzone.steps == 5, error
        assert my_gpuzone.sensitivity == 2, error
        assert my_gpuzone.polling == 2, error
        assert my_gpuzone.min_temp == 40, error
        assert my_gpuzone.max_temp == 70, error
        assert my_gpuzone.min_level == 35, error
        assert my_gpuzone.max_level == 100, error
        assert my_gpuzone.gpu_device_ids == [0], error
        assert my_gpuzone.nvidia_smi_path == '/usr/bin/nvidia-smi', error

    @pytest.mark.parametrize("device_ids, error", [
        # gpu_device_ids= invalid value(s)
        ('#, 0, 1',     'GpuZone.__init__() 5'),
        ('-1, 0, 1',    'GpuZone.__init__() 6'),
        ('0, 101, 1',   'GpuZone.__init__() 7')
    ])
    def test_init_n1(self, mocker: MockerFixture, device_ids: str, error: str):
        """Negative unit test for GpuZone.__init__() method. It contains the following steps:
            - mock print(), pyudev.Devices.from_device_file(), pyudev.Device, smfc.FanController.get_hwmon_path()
            - initialize a Config, Log, Ipmi, and GpuZone classes
            - ASSERT: if no assertion is raised for invalid values at initialization
        """
        my_td = TestData()
        count=3
        nvidia_smi_cmd = my_td.create_nvidia_smi_command(count)
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        my_config = ConfigParser()
        my_config[GpuZone.CS_GPU_ZONE] = {
            GpuZone.CV_GPU_ZONE_ENABLED: '1',
            GpuZone.CV_GPU_ZONE_GPU_IDS: device_ids,
            GpuZone.CV_GPU_ZONE_NVIDIA_SMI_PATH: nvidia_smi_cmd
        }
        my_log = Log(Log.LOG_DEBUG, Log.LOG_STDOUT)
        my_ipmi = Ipmi.__new__(Ipmi)
        with pytest.raises(Exception) as cm:
            GpuZone(my_log, my_ipmi, my_config)
        assert cm.type is ValueError, error
        del my_td

    #pylint: disable=protected-access
    @pytest.mark.parametrize("args, error", [
        (['--query-gpu=temperature.gpu', '--format=csv,noheader,nounits'],  'GpuZone._exec_nvidia_smi() 1'),
        (['0', '1', '2', '3'],                                              'GpuZone._exec_nvidia_smi() 2')
    ])
    def test_exec_nvidia_smi_p(self, mocker: MockerFixture, args:List[str], error: str):
        """Positive unit test for GpuZone._exec_nvidia_smi() method. It contains the following steps:
            - mock subprocess.run() function
            - initialize an empty GpuZone class
            - call GpuZone._exec_nvidia_smi() method
            - ASSERT: if subprocess.run() called with different parameters from specified argument list
        """
        expected_args: List[str]

        my_gpuzone = GpuZone.__new__(GpuZone)
        my_gpuzone.nvidia_smi_path = 'nvidia-smi'
        mock_subprocess_run = MagicMock()
        mock_subprocess_run.return_value=subprocess.CompletedProcess([], returncode=0, stdout='40', stderr='')
        mocker.patch('subprocess.run', mock_subprocess_run)
        my_gpuzone._exec_nvidia_smi(args)
        expected_args = [my_gpuzone.nvidia_smi_path]
        expected_args.extend(args)
        mock_subprocess_run.assert_called_with(expected_args, capture_output=True, check=False, text=True)
        assert mock_subprocess_run.call_count == 1, error

    #pylint: disable=R0801
    @pytest.mark.parametrize("nvidia_smi_path, exception, error", [
        # The real subprocess.run() executed (without sudo)
        ('/nonexistent/command', FileNotFoundError, 'GpuZone._exec_nvidia_smi() 3')
    ])
    def test_exec_nvidia_smi_n(self, nvidia_smi_path: str, exception: Any, error: str):
        """Negative unit test for GpuZone._exec_nvidia_smi() method. It contains the following steps:
            - mock subprocess.run() function if needed
            - initialize an empty GpuZone class
            - call GpuZone._exec_nvidia_smi() method
            - ASSERT: if no assertion was raised
        """
        my_gpuzone = GpuZone.__new__(GpuZone)
        my_gpuzone.nvidia_smi_path = nvidia_smi_path
        with pytest.raises(Exception) as cm:
            my_gpuzone._exec_nvidia_smi(['0', '1'])
        assert cm.type == exception, error
    # pylint: enable=R0801

    @pytest.mark.parametrize("count, temperatures, error", [
        (1, [32],                               'GpuZone._get_nth_temp() 1'),
        (2, [33, 34],                           'GpuZone._get_nth_temp() 2'),
        (4, [33, 34, 35, 38],                   'GpuZone._get_nth_temp() 3'),
        (8, [33, 34, 35, 38, 36, 37, 31, 30],   'GpuZone._get_nth_temp() 4')
    ])
    def test_get_nth_temp_p1(self, mocker: MockerFixture, count: int, temperatures: List[float], error: str):
        """Positive unit test for GpuZone._get_nth_temp() method. It contains the following steps:
            - mock print() function
            - initialize an empty GpuZone class
            - ASSERT: if the read temperature is different from the expected one
        """
        my_td = TestData()
        nvidia_smi_cmd = my_td.create_nvidia_smi_command(count, temperatures)
        my_gpuzone = GpuZone.__new__(GpuZone)
        my_gpuzone.gpu_device_ids = list(range(count))
        my_gpuzone.polling=2
        my_gpuzone.nvidia_smi_called=0
        my_gpuzone.nvidia_smi_path=nvidia_smi_cmd
        mock_print = MagicMock()
        mocker.patch('builtins.print', mock_print)
        for i in range(count):
            temp = my_gpuzone._get_nth_temp(i)
            assert temp == temperatures[i], error
        del my_td


# End.
